import logging
import logging.config
import os

import expan.core.debugging as dbgcls

reload(dbgcls)


def test_default_debugging(capsys):
	"""
	Test the debugging object created with default options.
	"""
	dbg = dbgcls.Dbg()
	dbg(-2, 'This should be an error')
	out, err = capsys.readouterr()
	print
	'out is: ' + out
	print
	'err is: ' + err
	errparts = err.split('|')
	print
	'errparts[1] is: [{}]'.format(errparts[1])
	assert (errparts[1] == 'test_default_debugging: This should be an error\n')

	dbg(-1, 'This should be a warning')
	dbg(0, 'This should be a stdout')
	# TODO: capsys appends output to existing, so we can't test each call's output individually.
	# out,err = capsys.readouterr()
	# assert(err.split('|')[1] == 'test_default_debugging: This should be a stdout')
	dbg(1, 'This should be a debug message')
	dbg(2, 'This should be a level 2 dbg message!')

	dbg.out(1, 'can also use like this!')


def test_debugging_with_file_config():
	"""
	Test the debugging object created with options from logging config file.
	TODO: test the output!
	"""
	logging.config.fileConfig(os.getcwd() + '/tests/tests_core/test-logging.conf')
	logger = logging.getLogger('simple')
	dbgf = dbgcls.Dbg(logger)

	dbgf(-2, 'This should be an error')
	dbgf(-1, 'This should be a warning')
	dbgf(0, 'This should be a stdout')
	dbgf(1, 'This should be a debug message')
	dbgf(2, 'This should be a level 2 dbg message!')

	dbgf.out(1, 'can also use like this!')


if __name__ == "__main__":
	pass  # TODO: call py.test?



import unittest
import warnings

# The simplefilter(always) stupidly does not clear the existing registry for the
# warnings, meaning that if any particular warning has already been masked by
# the 'once' filter, it will continue to be masked
# see https://docs.python.org/2/library/warnings.html#testing-warnings
warnings.simplefilter('always')

from expan.core.binning import *


class UtilTestCase(unittest.TestCase):
	"""
	Defines the setUp() and tearDown() functions for the statistics test cases.
	"""

	def setUp(self):
		"""
		Load the needed datasets for all UtilTestCases and set the random
		seed so that randomized algorithms show deterministic behaviour.
		"""
		np.random.seed(0)

	def tearDown(self):
		"""
		Clean up after the test
		"""
		# TODO: find out if we have to remove data manually
		pass


class BinningTestCase(UtilTestCase):
	"""
	Test cases for the binning function in core.
	"""

	def test_invalid_input(self):
		"""
		Check if errors are raised for invalid input (empty data and n_bins=0)
		"""

		# Check if error is raised for empty input
		with self.assertRaises(ValueError):
			create_binning(None, 10)

		with self.assertRaises(ValueError):
			create_binning(range(100), 0)

	def test_creation_range_100__nbins_1(self):
		"""
		Check if n_bins = 1 functions properly.
		"""
		# Create data
		x = range(100)
		# Calculate binning
		bins = create_binning(x, 1)
		r = bins.label(x, '{simplei}')
		# Expected result
		e = ['0_99']
		# Comparison
		self.assertEqual(r[0:1], e)

	def test_creation_range_100__nbins_2(self):
		"""
		Check if n_bins = 2 functions properly.
		"""
		# Create data
		x = range(100)
		# Calculate binning
		bins = create_binning(x, 2)
		r = bins.label(x, '{simplei}')
		# Expected result
		self.assertEqual(r[0], '0_50')
		self.assertEqual(r[-1], '50_99')

	def test_creation_range_100__n_bins_5(self):
		"""
		Check if n_bins = 5 functions properly.
		"""
		# Create data
		x = range(100)
		# Calculate binning
		bins = create_binning(x, 5)
		r = bins.label(x, '{simplei}')
		# Expected result
		# Comparison
		self.assertEqual(r[0], '0_20')
		self.assertEqual(r[20], '20_40')
		self.assertEqual(r[40], '40_60')
		self.assertEqual(r[60], '60_80')
		self.assertEqual(r[80], '80_99')
		self.assertEqual(r[99], '80_99')

	def test_creation_range_100__n_bins_8(self):
		"""
		Check if n_bins = 8 functions properly.
		"""
		# Create data
		x = range(100)
		# Calculate binning
		bins = create_binning(x, 8)
		labels = bins.label(x, '{simplei}')
		r = list(bins.labels('{simplei}'))
		r.sort()
		# Expected result
		e = ['88_99', '0_13', '63_76', '39_51', '51_63', '76_88', '26_39', '13_26']
		e.sort()
		# Comparison
		self.assertEqual(r, e)

	def test_creation_range_100__n_bins_10(self):
		"""
		Check if n_bins = 10 functions properly.
		"""
		# Create data
		x = range(100)
		# Calculate binning
		bins = create_binning(x, 10)
		r = bins.label(x, '{simplei}')

		# Expected result
		self.assertEqual(r[0], '0_10')
		self.assertEqual(r[10], '10_20')
		self.assertEqual(r[-1], '90_99')

	def test_creation_range_100__n_bins_20(self):
		"""
		Check if n_bins = 20 functions properly.
		"""
		# Create data
		x = range(100)

		# Calculate binning
		bins = create_binning(x, 20)
		r = bins.label(x, '{simplei}')

		# Expected result
		self.assertEqual(r[0], '0_5')
		self.assertEqual(r[6], '5_10')
		self.assertEqual(r[-1], '95_99')

	def test_creation_more_bins_than_data(self):
		"Check warnings and result if data has insufficient values for bins"
		# Create data
		x = [0] * 100 + [1] * 10
		# Calculate binning
		w = None
		with warnings.catch_warnings(record=True) as w:
			bins = create_binning(x, 3)

			self.assertEqual(len(w), 1)
			self.assertTrue(issubclass(w[-1].category, UserWarning))
			self.assertTrue('distinct values' in str(w[-1].message))

		r = list(bins.labels('{standard}'))
		r.sort()
		# Expected result
		e = ['[0.0,0.0]', '[1.0,1.0]']
		e.sort()
		# Comparison
		self.assertEqual(r, e)

	def test_creation_single_bin(self):
		"Check result if data has single value and request single bin"
		# Create data
		x = [0] * 100
		# Calculate binning
		bins = create_binning(x, 1)
		r = list(bins.labels())
		r.sort()
		# Expected result
		e = ['[0.0,0.0]']
		e.sort()
		# Comparison
		self.assertEqual(r, e)

	def test_creation_very_skewed_data(self):
		# Create data
		x = [0] * 10000 + range(300)
		# Calculate binning
		with warnings.catch_warnings(record=True) as w:
			bins = create_binning(x=x, nbins=4)
		r = bins.labels().tolist()
		# r.sort()
		# Expected result
		# e = ['0_1','1_100','100_200','200_300'] # DEFINE EXPECTED RESULT
		e = ['[0.0,0.0]', '[1.0,101.0)', '[101.0,200.0)', '[200.0,299.0]']
		# e.sort()
		# Comparison
		self.assertEqual(r, e)

	# How about this:
	# zero-inflated distribution
	# >>> dat = np.concatenate((np.random.poisson(3,1000), np.zeros(1000), np.repeat(np.nan,100)))

	# >>> interval_dict = get_binning(dat, 5)
	# >>> len(interval_dict)
	# 6

	# >>> dat = np.random.choice(['a','b','c','d','e'], 1000)
	# >>> len(get_binning(dat))
	# 5
	# >>> print set_binning(dat, interval_dict)
	# ['0.0_2.0' '2.0_3.0' '3.0_4.0' ..., '0.0_2.0' '0.0_2.0' '0.0_2.0']

	def test_creation_two_big_bins_noise_between(self):
		# Create data
		x = [0] * 10000 + range(300) + [301] * 10000
		# Calculate binning

		with warnings.catch_warnings(record=True) as w:
			bins = create_binning(x=x, nbins=10)
			self.assertEqual(len(w), 1)
			self.assertTrue(issubclass(w[-1].category, UserWarning))
			self.assertTrue('less bins than requested' in str(w[-1].message).lower())
		r = bins.labels().tolist()
		# r.sort()
		# Expected result
		# e = ['0_1','1_100','100_200','200_300','301_301'] # DEFINE EXPECTED RESULT
		e = ['[0.0,0.0]', '[1.0,301.0)', '[301.0,301.0]']
		# e.sort()
		# Comparison
		self.assertEqual(r, e)


class NumericalBinningClassTestCase(UtilTestCase):
	def test_manual_creation(self):
		"Test manual creation of numerical binning"
		bb = NumericalBinning()

		bb.lowers = [0, 1, 2, 3]
		bb.uppers = [1, 2, 3, 4]
		bb.lo_closed = [True] * 4
		bb.up_closed = [False] * 4
		bb.up_closed[3] = True

		n = np.nan

		with warnings.catch_warnings(record=True) as w:
			# ignore the invalid value warnings from the nan
			warnings.simplefilter('ignore', category=RuntimeWarning)
			res = bb._apply([-1, 0, 1, 2, 3, 4, 5, n, 1.99])

		np.testing.assert_array_equal(res, [-1, 0, 1, 2, 3, 3, -1, -1, 1])

	def test_manual_nan_bin(self):
		"Test handling of nan bin in numerical binning (manual creation)"
		bins = NumericalBinning()

		n = np.nan
		bins.lowers = [0, 1, 2, n]
		bins.uppers = [1, 2, 3, n]
		bins.lo_closed = [True] * 4
		bins.up_closed = [False] * 4

		# with warnings.catch_warnings(record=True) as w:
		# ignore the invalid value warnings from the nan
		#	warnings.simplefilter('ignore', category=RuntimeWarning)

		res = bins._apply([-1, 0, 1, 2, 3, n, 1.99, 2.99999])
		np.testing.assert_array_equal(res, [-1, 0, 1, 2, -1, 3, 1, 2])

	def test_binning(self):
		"Test various functions of numerical binning"
		values = np.arange(1000)
		nbins = 10

		bins = create_binning(values, nbins)
		self.assertTrue(isinstance(bins, NumericalBinning))
		self.assertEqual(len(bins), nbins)

		# Test using midpoint as label
		# This allows the label to be passed back into the binning.label
		# and will label identically
		labels = bins.mid(values)
		self.assertEqual(labels[0], 50.)
		self.assertEqual(labels[99], 50.)
		self.assertEqual(labels[100], 150.)

		# Test using midpoint as a formatstr
		labels = bins.label(values, '{mid:.2f}')
		self.assertEqual(labels[0], '50.00')
		self.assertEqual(labels[99], '50.00')

		labels = bins.label(values, '{set_notation}')
		self.assertEqual(labels[0], '[0.0,100.0)')
		self.assertEqual(labels[-1], '[900.0,999.0]')

		# Test a certain label format...
		labels = bins.label(values, '({lo:.0f}{lo_cond}x{up_cond}{up:.0f})')
		self.assertEqual(labels[0], '(0<=x<100)')
		self.assertEqual(labels[-1], '(900<=x<=999)')

		# Test a certain label format with the shortcuts...
		labels = bins.label(values, '({conditions})')
		self.assertEqual(labels[0], '(0.0<=x<100.0)')
		self.assertEqual(labels[-1], '(900.0<=x<=999.0)')

		# See if we can format the numbers how we like
		labels = bins.label(values, '({lo:.2f}{lo_cond}x{up_cond}{up:.2f})')
		self.assertEqual(labels[0], '(0.00<=x<100.00)')
		self.assertEqual(labels[-1], '(900.00<=x<=999.00)')

		# Test a certain label format...
		labels = bins.label(values, '{iter.uppercase}{lo_bracket}{lo:.0f},{up:.0f}{up_bracket}')
		self.assertEqual(len(bins), nbins)
		self.assertEqual(labels[0], 'A[0,100)')
		self.assertEqual(labels[99], 'A[0,100)')
		self.assertEqual(labels[100], 'B[100,200)')
		self.assertEqual(labels[-1], 'J[900,999]')

	def test_unseen_data(self):
		"Test unseen data with numerical binning"
		seen = np.arange(1000.)
		nbins = 10
		unseen = np.arange(1100)

		bins = create_binning(seen, nbins)
		self.assertTrue(isinstance(bins, NumericalBinning))
		self.assertEqual(len(bins), nbins)

		labels = bins.mid(unseen)
		self.assertEqual(labels[0], 50.)
		self.assertEqual(labels[99], 50.)
		self.assertEqual(labels[100], 150.)
		self.assertEqual(labels[1000 - 1], 949.5)  # last bin is closed-closed

		# Test that the unseen data is given NaN
		np.testing.assert_equal(labels[1000], np.nan)  # tests equal

		labels = bins.label(unseen, '({conditions})')
		self.assertEqual(labels[0], '(0.0<=x<100.0)')
		self.assertEqual(labels[-1], '(unseen)')
		# This would be cool: (would require knowing that bins span space)
		# self.assertEqual(labels[-1], '(x<0.0 or x>999.0)')

		labels = bins.label(unseen)
		self.assertEqual(labels[-1], '[unseen]')

		labels = bins.label(unseen, '{iter.uppercase} ({conditions})')
		self.assertEqual(labels[0], 'A (0.0<=x<100.0)')
		self.assertEqual(labels[100], 'B (100.0<=x<200.0)')
		self.assertEqual(labels[-1], '? (unseen)')

		labels = bins.label(unseen, '{iter.uppercase}')
		self.assertEqual(labels[0], 'A')
		self.assertEqual(labels[100], 'B')
		self.assertEqual(labels[-1], '?')

	def test_numerical_nans(self):
		"Test handling of nan in numerical binning"
		n = np.nan
		values = np.arange(1002.)
		values[-2] = n
		values[-1] = n
		nbins = 10

		bins = create_binning(values, nbins)
		self.assertTrue(isinstance(bins, NumericalBinning))

		# we say that nan bins are e
		self.assertEqual(len(bins), nbins + 1)

		labels = bins.label(values)
		self.assertEqual(labels[-2], '[nan,nan]')
		self.assertEqual(labels[-1], '[nan,nan]')
		self.assertEqual(labels[-3], '[900.0,999.0]')
		self.assertEqual(labels[0], '[0.0,100.0)')

		# Test using midpoint as label
		# This allows the label to be passed back into the binning.label
		# and will label identically
		labels = bins.mid(values)
		self.assertTrue(np.isnan(labels[-2]))
		self.assertTrue(np.isnan(labels[-1]))
		self.assertEqual(labels[-3], 949.5)
		self.assertEqual(labels[0], 50.)
		self.assertEqual(labels[99], 50.)
		self.assertEqual(labels[100], 150.)

		# test unseen
		values[-3] = 1001.
		labels = bins.label(values, format_str='{iter.integer} ({conditions})')
		self.assertEqual(labels[-3], '? (unseen)')

		# following assumes the nan bin will be the first bin...
		# this may be an assumption we want to change.
		self.assertEqual(labels[0], '1 (0.0<=x<100.0)')


class CategoricalBinningClassTestCase(UtilTestCase):
	def test_manual_creation(self):
		"Test manual creation of categorical binning"
		bb = CategoricalBinning()
		n = np.nan

		bb.categories += [['a', 'b']]
		bb.categories += [['c', 'd', 'e']]
		bb.categories += [['f', 'g']]
		bb.categories += [['h']]

		res = bb._apply(['z', 'a', 'c', 'f', 'h', n, 'x', 0, 'aa'])

		np.testing.assert_array_equal(res, [-1, 0, 1, 2, 3, -1, -1, -1, -1])

	def test_manual_nan_bin(self):
		"Test handling of nan bin in categorical binning (manual creation)"
		bins = CategoricalBinning()
		n = np.nan

		bins.categories += [['a', 'b']]
		bins.categories += [['c', 'd', 'e']]
		bins.categories += [['f', 'g']]
		bins.categories += [['h', np.nan]]
		bins.categories += [['i', 'j']]

		res = bins._apply(['z', 'a', 'c', 'f', 'h', n, 'x', 0, 'aa', 'j'])
		np.testing.assert_array_equal(res, [-1, 0, 1, 2, 3, 3, -1, -1, -1, 4])

		bins = CategoricalBinning()
		bins.categories += [['a', 'b']]
		bins.categories += [[n]]
		res = bins._apply(['a', 'b', 'c', n])
		np.testing.assert_array_equal(res, [0, 0, -1, 1])

		x = ['A'] * 50 + ['B'] * 10 + ['C'] * 20 + [np.nan] * 10
		bins = create_binning(x=x, nbins=3)
		r = bins.labels().tolist()
		e = ['{A}', '{B}', '{C}', '{nan}']
		self.assertEqual(r, e)

	def test_creation_categorical_data_enough_bins(self):
		"Test binning of normal categorical data"
		# Create data
		x = ['A'] * 50 + ['B'] * 10 + ['C'] * 20

		# Calculate binning
		with warnings.catch_warnings(record=True) as w:
			bins = create_binning(x=x, nbins=10)

			self.assertEqual(len(w), 1)
			self.assertTrue(issubclass(w[-1].category, UserWarning))
			self.assertTrue(
				'insufficient distinct values' in str(w[-1].message).lower())

		self.assertTrue(isinstance(bins, CategoricalBinning))
		self.assertEqual(len(bins), 3)

		# Test the labels of the bins themselves.
		r = bins.label(x)
		self.assertEqual(r[0], '{A}')
		self.assertEqual(r[50], '{B}')
		self.assertEqual(r[60], '{C}')
		self.assertEqual(r[-1], '{C}')

	def test_creation_categorical_data_less_bins(self):
		"Test binning of normal categorical data"

		x = ['A'] * 50 + ['B'] * 10 + ['C'] * 20

		bins = create_binning(x=x, nbins=2)
		self.assertEqual(len(bins), 2)

		r = bins.label(x)
		self.assertEqual(r[0], '{A}')
		self.assertEqual(r[50], '{B,C}')
		self.assertEqual(r[60], '{B,C}')
		self.assertEqual(r[-1], '{B,C}')

	def test_categorical_with_unseen_data(self):
		"Test categorical binning with unseen data"

		x = ['A'] * 50 + ['B'] * 10 + ['C'] * 20
		y = ['A'] * 10 + ['B'] * 10 + ['C'] * 10 + ['D'] * 10 + ['A,B'] * 10 + ['AB'] * 10

		bins = create_binning(x=x, nbins=3)
		self.assertEqual(len(bins), 3)

		r = bins.label(y)
		self.assertEqual(r[0], '{A}')
		self.assertEqual(r[10], '{B}')
		self.assertEqual(r[20], '{C}')
		# now unseen:
		self.assertEqual(r[30], '{unseen}')
		for rr in r[31:]:
			self.assertEqual(rr, '{unseen}')

		r = bins.label(y, '{iter.integer}: {standard}')
		self.assertEqual(r[0], '0: {A}')
		self.assertEqual(r[-1], '?: {unseen}')

	def test_categorical_order(self):
		"Test that categorical bins requiring no merging are created in alphabetical order"
		x = ['C'] * 50 + ['B'] * 10 + ['A'] * 20
		bins = create_binning(x=x, nbins=3)
		r = bins.label(x, '{iter.integer}: {standard}')
		self.assertEqual(r[0], '2: {C}')
		self.assertEqual(r[50], '1: {B}')
		self.assertEqual(r[60], '0: {A}')



import warnings

import numpy as np
import pandas as pd
from scipy import stats


def _delta_mean(x, y):
	"Implemented as function to allow calling from bootstrap."
	return np.nanmean(x) - np.nanmean(y)


def delta(x, y, assume_normal=True, percentiles=[2.5, 97.5],
		  min_observations=20, nruns=10000, relative=False):
	"""
	Calculates the difference of means between the samples (x-y) in a
	statistical sense, i.e. with confidence intervals.

	NaNs are ignored: treated as if they weren't included at all. This is done
	because at this level we cannot determine what a NaN means. In some cases,
	a NaN represents missing data that should be completely ignored, and in some
	cases it represents inapplicable (like PCII for non-ordering customers) - in
	which case the NaNs should be replaced by zeros at a higher level. Replacing
	with zeros, however, would be completely incorrect for return rates.

	Computation is done in form of treatment minus control, i.e. x-y

	Args:
		x (array_like): sample of a treatment group
		y (array_like): sample of a control group
		assume_normal (boolean): specifies whether normal distribution
			assumptions can be made
		percentiles (list): list of percentile values for confidence bounds
		min_observations (integer): minimum number of observations needed
		nruns (integer): only used if assume normal is false
		relative (boolean): if relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In	this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		float: mean value of the difference
		dict: percentile levels (index) and values
		int: size of x excluding NA values
		int: size of y excluding NA values
		float: absolute mean of x
		float: absolute mean of y
	"""
	# Checking if data was provided
	if x is None or y is None:
		raise ValueError('Please provide two non-None samples.')

	# Coercing missing values to right format
	_x = np.array(x, dtype=float)
	_y = np.array(y, dtype=float)

	x_nan = np.isnan(_x).sum()
	y_nan = np.isnan(_y).sum()
	if x_nan > 0:
		warnings.warn('Discarding ' + str(x_nan) + ' NaN(s) in the x array!')
	if y_nan > 0:
		warnings.warn('Discarding ' + str(y_nan) + ' NaN(s) in the y array!')

	ss_x = sample_size(_x)
	ss_y = sample_size(_y)

	# Checking if enough observations are left after dropping NaNs
	if min(ss_x, ss_y) < min_observations:
		# Set mean to nan
		mu = np.nan
		# Create nan dictionary
		c_i = dict(zip(percentiles, np.empty(len(percentiles)) * np.nan))
	else:
		# Computing the mean
		mu = _delta_mean(_x, _y)
		# Computing the confidence intervals
		if assume_normal:
			c_i = normal_sample_difference(x=_x, y=_y, percentiles=percentiles,
										   relative=relative)
		else:
			c_i, _ = bootstrap(x=_x, y=_y, percentiles=percentiles, nruns=nruns,
							   relative=relative)

	# Return the result structure
	return mu, c_i, ss_x, ss_y, np.nanmean(_x), np.nanmean(_y)


def sample_size(x):
	"""
	Calculates sample size of a sample x
	Args:
	    x (array_like): sample to calculate sample size

	Returns:
	    int: sample size of the sample excluding nans

	"""
	# cast into a dummy numpy array to infer the dtype
	if ~isinstance(x, np.ndarray):
		dummy = np.array(x)
	is_numeric = np.issubdtype(dummy.dtype, np.number)

	if is_numeric:
		# Coercing missing values to right format
		_x = np.array(x, dtype=float)
		x_nan = np.isnan(_x).sum()

	# assuming categorical sample
	elif isinstance(x, pd.core.series.Series):
		x_nan = x.str.contains('NA').sum()
	else:
		x_nan = list(x).count('NA')

	if x_nan > 0:
		warnings.warn('Discarding ' + str(x_nan) + ' NaN(s) in the x array!')

	return len(x) - x_nan


def chi_square(x, y, min_counts=5):
	"""
	Performs the chi-square homogeneity test on categorical arrays x and y

	Args:
		x (array_like): sample of the treatment variable to check
		y (array_like): sample of the control variable to check
		min_counts (int): drop categories where minimum number of observations
		                or expected observations is below min_counts for x or y

	Returns:
		float: p-value
		float: chi-square value
		int: number of attributes used (after dropping)

	"""
	# Checking if data was provided
	if x is None or y is None:
		raise ValueError('Please provide two samples.')

	# Check if data is not empty
	if not len(x) or not len(y):
		return np.nan

	# Transform input to categorical variable
	_x = pd.Categorical(x)
	_y = pd.Categorical(y)

	#
	treat_counts = _x.value_counts()
	control_counts = _y.value_counts()
	# Get observed counts for both _x and _y for each category
	# (=contingency table) and set the counts for non occuring categories to 0
	observed_ct = pd.DataFrame([treat_counts, control_counts]).fillna(0)
	# Ensure at least a frequency of 5 at every location in observed_ct,
	# otherwise drop categorie see
	# http://docs.scipy.org/doc/scipy-0.16.1/reference/generated/scipy.stats.chisquare.html
	observed_freqs = observed_ct[observed_ct >= min_counts].dropna(axis=1)

	# Calculate expected counts for chi-square homogeneity test
	# expected_freqs = group_totals*category_totals/all_totals
	# see e.g. Fahrmeir, L., Kuenstler, R., Pigeot, I., & Tutz, G. (2007).
	#          Statistik: Der Weg zur Datenanalyse. Springer-Verlag.
	all_totals = observed_freqs.sum().sum()
	category_totals = observed_freqs.sum(axis=0)
	expected_freqs = np.outer(category_totals,
							  observed_freqs.sum(axis=1) / all_totals).T

	# The actual degrees of freedom for the test are dof=(num_categories-1)
	# however the chisquare() function assumes dof=k-1, with
	# k = num_variants*num_categories
	# see http://docs.scipy.org/doc/scipy-0.16.1/reference/generated/scipy.stats.chisquare.html
	# Therefore, we have to correct using
	# ddof=(2*num_categories-1) - num_categories-1
	# Calculate the chi-square statistic and p-value
	es = expected_freqs.shape
	delta_dof = (es[0] * es[1] - 1) - (es[0] - 1) * (es[1] - 1)
	chisqr, p_val = stats.chisquare(f_obs=observed_freqs,
									f_exp=expected_freqs,
									ddof=delta_dof,
									axis=None)
	# Return the p-value
	return p_val, chisqr, es[1]


def alpha_to_percentiles(alpha):
	"""
	Transforms alpha value to corresponding percentile.

	Args:
		alpha (float): alpha values to transform

	Returns:
		list of percentiles corresponding to given alpha
	"""
	# Compute the percentiles
	return [100. * alpha / 2, 100. * (1 - (alpha / 2))]


def bootstrap(x, y, func=_delta_mean, nruns=10000, percentiles=[2.5, 97.5],
			  min_observations=20, return_bootstraps=False, relative=False):
	"""
	Bootstraps the Confidence Intervals for a particular function comparing
	two samples. NaNs are ignored (discarded before calculation).

	Args:
		x (array like): sample of treatment group
		y (array like): sample of control group
		func (function): function of which the distribution is to be computed.
			The default comparison metric is the difference of means. For
			bootstraping correlation: func=lambda x,y: np.stats.pearsonr(x,y)[0]
		nruns (integer): number of bootstrap runs to perform
		percentiles (list):	The values corresponding to the given percentiles
			are returned. The default percentiles (2.5% and 97.5%) correspond to
			an alpha of 0.05.
		min_observations (integer): minimum number of observations necessary
		return_bootstraps (boolean): If this variable is set the bootstrap sets
			are returned otherwise the first return value is empty.
		relative (boolean): if relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In	this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		dict: percentile levels (index) and values
		np.array (nruns): array containing the bootstraping results per run
	"""
	# Checking if data was provided
	if x is None or y is None:
		raise ValueError('Please provide two non-None samples.')

	# Transform data to appropriate format
	_x = np.array(x, dtype=float)
	_y = np.array(y, dtype=float)
	ss_x = _x.size - np.isnan(_x).sum()
	ss_y = _y.size - np.isnan(_y).sum()

	# Checking if enough observations are left after dropping NaNs
	if min(ss_x, ss_y) < min_observations:
		# Create nan percentile dictionary
		c_val = dict(zip(percentiles, np.empty(len(percentiles)) * np.nan))
		return (c_val, None)
	else:
		# Initializing bootstraps array and random sampling for each run
		bootstraps = np.ones(nruns) * np.nan
		for run in range(nruns):
			# Randomly chose values from _x and _y with replacement
			xp = _x[np.random.randint(0, len(_x), size=(len(_x),))]
			yp = _y[np.random.randint(0, len(_y), size=(len(_y),))]
			# Application of the given function to the bootstraps
			bootstraps[run] = func(xp, yp)
		# If relative is set subtract mean from bootstraps
		if relative:
			bootstraps -= np.nanmean(bootstraps)
		# Confidence values per given percentile as dictionary
		c_val = dict(zip(percentiles, np.percentile(bootstraps, q=percentiles)))
		return (c_val, None) if not return_bootstraps else (c_val, bootstraps)


def pooled_std(std1, n1, std2, n2):
	"""
	Returns the pooled estimate of standard deviation. Assumes that population
	variances are equal (std(v1)**2==std(v2)**2) - this assumption is checked
	for reasonableness and an exception is raised if this is strongly violated.

	Args:
		std1 (float): standard deviation of first sample
		n1 (integer): size of first sample
		std2 (float): standard deviation of second sample
		n2 (integer): size of second sample

	Returns:
		float: Pooled standard deviation

	For further information visit:
		http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html

	TODO: also implement a version for unequal variances
	"""
	if not (0.5 < (std1 ** 2) / (std2 ** 2) < 2.):
		warnings.warn('Sample variances differ too much to assume that '
					  'population variances are equal.')

	return np.sqrt(((n1 - 1) * std1 ** 2 + (n2 - 1) * std2 ** 2) / (n1 + n2 - 2))


def normal_percentiles(mean, std, n, percentiles=[2.5, 97.5], relative=False):
	"""
	Calculate the percentile values for a normal distribution with parameters
	estimated from samples.

	Args:
		mean (float): mean value of the distribution
		std (float): standard deviation of the distribution
		n (integer): number of samples
		percentiles (list): list of percentile values to compute
		relative (boolean): if relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		dict: percentiles and corresponding values

	For more information visit:
		http://www.itl.nist.gov/div898/handbook/eda/section3/eda352.htm
		http://www.boost.org/doc/libs/1_46_1/libs/math/doc/sf_and_dist/html/math_toolkit/dist/stat_tut/weg/st_eg/tut_mean_intervals.html
		http://www.stat.yale.edu/Courses/1997-98/101/confint.htm
	"""
	# Computing standard error
	st_error = std / np.sqrt(n)

	# Mapping percentiles via standard error
	if relative:
		return dict([(p, stats.t.ppf(p / 100.0, df=n - 1) * st_error)
					 for p in percentiles])
	else:
		return dict([(p, mean + stats.t.ppf(p / 100.0, df=n - 1) * st_error)
					 for p in percentiles])


def normal_sample_percentiles(values, percentiles=[2.5, 97.5], relative=False):
	"""
	Calculate the percentile values for a sample assumed to be normally
	distributed. If normality can not be assumed, use bootstrap_ci instead.
	NaNs are ignored (discarded before calculation).

	Args:
		values (array-like): sample for which the normal distribution
			percentiles are computed.
		percentiles (list): list of percentile values to compute
		relative (boolean): if relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In	this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		dict: percentiles and corresponding values

	For further information visit:
		http://www.itl.nist.gov/div898/handbook/eda/section3/eda352.htm
		http://www.boost.org/doc/libs/1_46_1/libs/math/doc/sf_and_dist/html/math_toolkit/dist/stat_tut/weg/st_eg/tut_mean_intervals.html
		http://www.stat.yale.edu/Courses/1997-98/101/confint.htm
	"""
	# Coerce data to right format
	_x = np.array(values, dtype=float)
	_x = _x[~np.isnan(_x)]

	# Determine distribution parameters
	mean = np.mean(_x)
	std = np.std(_x)
	n = len(_x)

	# Actual computation is done in normal_percentiles
	return normal_percentiles(mean=mean, std=std, n=n, percentiles=percentiles,
							  relative=relative)


def normal_sample_difference(x, y, percentiles=[2.5, 97.5], relative=False):
	"""
	Calculates the difference distribution of two normal distributions given
	by their samples.

	Computation is done in form of treatment minus control. It is assumed that
	the standard deviations of both distributions do not differ too much.

	Args:
		x (array-like): sample of a treatment group
		y (array-like): sample of a control group
		percentiles (list): list of percentile values to compute
		relative (boolean): If relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		dict: percentiles and corresponding values
	"""
	# Coerce data to right format
	_x = np.array(x, dtype=float)
	_x = _x[~np.isnan(_x)]
	_y = np.array(y, dtype=float)
	_y = _y[~np.isnan(_y)]

	# Calculate statistics
	mean1 = np.mean(_x)
	mean2 = np.mean(_y)
	std1 = np.std(_x)
	std2 = np.std(_y)
	n1 = len(_x)
	n2 = len(_y)

	# Push calculation to normal difference function
	return normal_difference(mean1=mean1, std1=std1, n1=n1, mean2=mean2,
							 std2=std2, n2=n2, percentiles=percentiles, relative=relative)


def normal_difference(mean1, std1, n1, mean2, std2, n2, percentiles=[2.5, 97.5],
					  relative=False):
	"""
	Calculates the difference distribution of two normal distributions.

	Computation is done in form of treatment minus control. It is assumed that
	the standard deviations of both distributions do not differ too much.

	Args:
		mean1 (float): mean value of the treatment distribution
		std1 (float): standard deviation of the treatment distribution
		n1 (integer): number of samples of the treatment distribution
		mean2 (float): mean value of the control distribution
		std2 (float): standard deviation of the control distribution
		n2 (integer): number of samples of the control distribution
		percentiles (list): list of percentile values to compute
		relative (boolean): If relative==True, then the values will be returned
			as distances below and above the mean, respectively, rather than the
			absolute values. In	this case, the interval is mean-ret_val[0] to
			mean+ret_val[1]. This is more useful in many situations because it
			corresponds with the sem() and std() functions.

	Returns:
		dict: percentiles and corresponding values

	For further information vistit:
			http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html
	"""
	# TODO: Figure out how to pass directly to normal_percentiles()

	# Compute combined parameters from individual parameters
	mean = mean1 - mean2
	std = pooled_std(std1, n1, std2, n2)
	# Computing standard error
	st_error = std * np.sqrt(1. / n1 + 1. / n2)
	# Computing degrees of freedom
	d_free = n1 + n2 - 2

	# Mapping percentiles via standard error
	if relative:
		return dict([(p, stats.t.ppf(p / 100.0, df=d_free) * st_error)
					 for p in percentiles])
	else:
		return dict([(p, mean + stats.t.ppf(p / 100.0, df=d_free) * st_error)
					 for p in percentiles])

def estimate_std(x, mu, pctile):
	"""Estimate the standard deviation from a given percentile, according to
	the z-score:
		z = (x - mu) / sigma

	Args:
		x (float): cumulated density at the given percentile
		mu (float): mean of the distribution
		pctile (float): percentile value (between 0 and 100)

	Returns:
		float: estimated standard deviation of the distribution
	"""
	return (x - mu) / stats.norm.ppf(pctile / 100.0) 
	

if __name__ == '__main__':
	# doctest.testmod()
	delta(None, None)



# TODO: question, shouldnt the module level functions be private (_*), because
# the proper interface is through the Experiment instance functions, surely?

# import numpy as np
import statistics as statx
import warnings

import binning as binmodule  # name conflict with binning...
import numpy as np
import pandas as pd
from experimentdata import ExperimentData
from results import Results, delta_to_dataframe_all_variants, feature_check_to_dataframe

# raise the same warning multiple times
warnings.simplefilter('always', UserWarning)

from debugging import Dbg

def _binned_deltas(df, variants, n_bins=4, binning=None, cumulative=False,
				   assume_normal=True, percentiles=[2.5, 97.5],
				   min_observations=20, nruns=10000, relative=False,
				   label_format_str='{standard}'):
	"""
	Calculates the feature dependent delta. Only used internally. All
	calculation by subgroup_delta() and time_dependant_delta() is pushed here.

	Args:
	  df (pandas DataFrame): 3 columns. The order of the columns is expected
	      to be variant, feature, kpi.
	  variants (list of 2): 2 entries, first entry is the treatment variant,
	      second entry specifies the baseline variant
	      TODO: currently only the baseline variant is extracted from this list
	        and deltas are calculated for all variants (see bug OCTO-869)

	  n_bins (integer): number of bins to create if binning is None
	  binning (list of bins): preset (if None then binning is created)

	  percentiles (list): list of percentile values to compute
	  min_observations (integer): minimum number of observations necessary. If
	      less observations are given, then NaN is returned.
	  nruns (integer): number of bootstrap runs to perform if assume normal is
	      set to False.
	  relative (boolean): If relative==True, then the values will be returned
	      as distances below and above the mean, respectively, rather than the
	      absolute values. In	this case, the interval is mean-ret_val[0] to
	      mean+ret_val[1]. This is more useful in many situations because it
	      corresponds with the sem() and std() functions.
	  label_format_str (string): format string for the binning label function.

	Returns:
	  pandas.DataFrame: bin-name, mean, percentile and corresponding values
	  list: binning used
	"""

	# Performing binning of feature on feat2
	if binning is None:
		binning = binmodule.create_binning(df.iloc[:, 1], nbins=n_bins)

	if cumulative==True and type(binning)!=binmodule.NumericalBinning:
		raise ValueError("Cannot calculate cumulative deltas for non-numerical binnings")

	# Applying binning to feat1 and feat2 arrays
	df.loc[:, '_tmp_bin_'] = binning.label(data=df.iloc[:, 1],
										   format_str=label_format_str)

	# Initialize result object as data frame with bin keys as index
	def do_delta(f, bin_name):
		# find the corresponding bin in the baseline variant
		baseline_metric = f.iloc[:, 2][(f.iloc[:, 0] == variants[1])]
		out_df=pd.DataFrame()

		for v in f['variant'].unique():
			v_metric =  f.iloc[:, 2][ (f.iloc[:,0]==v) ]
			df = delta_to_dataframe_all_variants(f.columns[2],
													*statx.delta(
														x=v_metric,
														y=baseline_metric,
														assume_normal=assume_normal,
														percentiles=percentiles,
														min_observations=min_observations,
														nruns=nruns, relative=relative))

			# add new index levels for variant and binning
			df['_tmp_bin_']=bin_name
			df['variant'] = v
			df.set_index(['variant', '_tmp_bin_'], append=True, inplace=True)
			df=df.reorder_levels(['variant', '_tmp_bin_', 'metric',
			                      'subgroup_metric', 'subgroup',
			                      'statistic', 'pctile'])

			out_df=out_df.append(df)
		return out_df

	# Actual calculation
	result = pd.DataFrame()
	unique_tmp_bins=df['_tmp_bin_'].unique()
	for bin in unique_tmp_bins:
		if not cumulative:
			result=result.append(do_delta(df[df['_tmp_bin_'] == bin], bin))
		else:
			result=result.append(do_delta(df[df['_tmp_bin_'] <= bin], bin))

	# unstack variant
	result = result.unstack(0)
	# drop _tmp_bin_ in the input data frame
	del df['_tmp_bin_']

	result.index = result.index.swaplevel(0, 2)
	result.index = result.index.swaplevel(0, 1)
	# Return result and binning
	return Results(result, {'binning': binning})


def _delta_all_variants(metric_df, baseline_variant, assume_normal=True,
						percentiles=[2.5, 97.5], min_observations=20,
						nruns=10000, relative=False):
	"""Applies delta to all variants, given a metric."""
	baseline_metric = metric_df.iloc[:, 2][metric_df.iloc[:, 1] == baseline_variant]
	do_delta = (lambda f: delta_to_dataframe_all_variants(f.columns[2],
														  *statx.delta(
															  x=f.iloc[:, 2],
															  y=baseline_metric,
															  assume_normal=assume_normal,
															  percentiles=percentiles,
															  min_observations=min_observations,
															  nruns=nruns, relative=relative)))
	# Actual calculation
	return metric_df.groupby('variant').apply(do_delta).unstack(0)


def _feature_check_all_variants(metric_df, baseline_variant, assume_normal=True,
								percentiles=[2.5, 97.5], min_observations=20,
								nruns=10000, relative=False):
	"""Applies delta to all variants, given a metric."""
	baseline_metric = metric_df.iloc[:, 2][metric_df.variant == baseline_variant]

	def do_delta_numerical(df):
		mu, ci, ss_x, ss_y, mean_x, mean_y = statx.delta(x=df.iloc[:, 2],
														 y=baseline_metric,
														 assume_normal=assume_normal,
														 percentiles=percentiles,
														 min_observations=min_observations,
														 nruns=nruns,
														 relative=relative)
		return feature_check_to_dataframe(metric=df.columns[2],
										  samplesize_variant=ss_x,
										  mu=mu,
										  pctiles=ci,
										  mu_variant=mean_x)

	def do_delta_categorical(df):
		pval = statx.chi_square(x=df.iloc[:, 2], y=baseline_metric)[0]
		ss_x = statx.sample_size(df.iloc[:, 2])
		return feature_check_to_dataframe(metric=df.columns[2],
										  samplesize_variant=ss_x,
										  pval=pval)

	# numerical feature
	if np.issubdtype(metric_df.iloc[:, 2].dtype, np.number):
		return metric_df.groupby('variant').apply(do_delta_numerical).unstack(0)
	# categorical feature
	else:
		return metric_df.groupby('variant').apply(do_delta_categorical).unstack(0)


def subgroup_deltas(df, variants, n_bins=4, binning=None,
					assume_normal=True, percentiles=[2.5, 97.5],
					min_observations=20, nruns=10000, relative=False):
	"""
	Calculates the feature dependent delta.

	Args:
	  df (pandas DataFrame): 3 columns. The order of the columns is expected
	      to be variant, feature, kpi.
	  variants (list of 2): 2 entries, first entry is the treatment variant,
	      second entry specifies the baseline variant

	  n_bins (integer): number of bins to create if binning is None
	  binning (list of bins): preset (if None then binning is created)

	  assume_normal (boolean): specifies whether normal distribution
	      assumptions can be made
	  percentiles (list): list of percentile values to compute
	  min_observations (integer): minimum number of observations necessary. If
	      less observations are given, then NaN is returned.
	  nruns (integer): number of bootstrap runs to perform if assume normal is
	      set to False.
	  relative (boolean): If relative==True, then the values will be returned
	      as distances below and above the mean, respectively, rather than the
	      absolute values. In	this case, the interval is mean-ret_val[0] to
	      mean+ret_val[1]. This is more useful in many situations because it
	      corresponds with the sem() and std() functions.

	Returns:
	  pandas.DataFrame: bin-name, mean, percentile and corresponding values
	  list: binning used
	"""

	# Push computation to _binned_deltas() function
	result = _binned_deltas(df=df, variants=variants, n_bins=n_bins,
							assume_normal=assume_normal, percentiles=percentiles,
							min_observations=min_observations, nruns=nruns,
							relative=relative)

	# TODO: Add binning to result metadata

	# Reformating of the index names in the result data frame object
	result.df.reset_index('subgroup', drop=True, inplace=True)
	result.df.index.set_names('subgroup', level=2, inplace=True)
	result.df.index.set_levels(levels=[df.columns[1]],
							   level='subgroup_metric', inplace=True)

	# Returning Result object containing result and the binning
	return result


def time_dependent_deltas(df, variants, time_step=1, cumulative=False,
						  assume_normal=True, percentiles=[2.5, 97.5],
						  min_observations=20, nruns=10000, relative=False):
	"""
	Calculates the time dependent delta.

	Args:
	  df (pandas DataFrame): 3 columns. The order of the columns is expected
	      to be variant, time, kpi.
	  variants (list of 2): 2 entries, first entry is the treatment variant,
	      second entry specifies the baseline variant

	  time_step (integer): time_step used for analysis.

	  percentiles (list): list of percentile values to compute
	  min_observations (integer): minimum number of observations necessary. If
	      less observations are given, then NaN is returned
	  nruns (integer): number of bootstrap runs to perform if assume normal is
	      set to False.
	  relative (boolean): If relative==True, then the values will be returned
	      as distances below and above the mean, respectively, rather than the
	      absolute values. In	this case, the interval is mean-ret_val[0] to
	      mean+ret_val[1]. This is more useful in many situations because it
	      corresponds with the sem() and std() functions.

	Returns:
	  pandas.DataFrame: bin-name, mean, percentile and corresponding values
	  list: binning used
	"""
	# TODO: allow times to have time stamp format
	# TODO: allow start time and end time format
	# TODO: fill with zeros

	# Create time binning with time_step
	#time_bin = (lambda x: round(x / float(time_step) + 0.5) * time_step)

	# Apply time binning vectorized to each element in the input array
	#df['_tmp_time_'] = df.iloc[:, 1].apply(time_bin)

	# Get appropriate bin number
	#n_bins = len(pd.unique(df['_tmp_time_']))

	# create binning manually, ASSUMING uniform sampling
	tpoints = np.unique(df.iloc[:,1])
	binning = binmodule.NumericalBinning(uppers=tpoints, lowers=tpoints, 
		up_closed=[True]*len(tpoints), lo_closed=[True]*len(tpoints))

	# Push computation to _binned_deltas() function
	result = _binned_deltas(df=df, variants=variants, binning=binning,
	                        cumulative=cumulative,
							assume_normal=assume_normal, percentiles=percentiles,
							min_observations=min_observations, nruns=nruns,
							relative=relative, label_format_str='{mid}')

	# Reformating of the index names in the result data frame object
	result.df.index.set_names('time', level=2, inplace=True)

	# Returning Result object containing result and the binning
	return result


################################################################################


# TODO: can we unbury this class a bit - be nice to have it as the first
# thing in the file?
# TODO: add filtering functionality: we should be able to operate on this
# class to exclude data points, and save all these operations in a log that then
# is preserved in all results.
class Experiment(ExperimentData):
	"""
	Class which adds the analysis functions to experimental data.
	"""

	# TODO: rearrange arguments
	# TODO: add a constructor that takes an ExperimentData!
	def __init__(self, baseline_variant, metrics_or_kpis, metadata={},
				 features='default', dbg=None):
		# Call constructor of super class
		super(Experiment, self).__init__(metrics_or_kpis, metadata, features)

		# If no baseline variant is found
		if ((baseline_variant not in self.kpis.index.levels[
			self.primary_indices.index('variant')])
			and (baseline_variant not in self.features.index.levels[
				self.primary_indices.index('variant')])):
			raise KeyError('baseline_variant ({}) not present in KPIs or features.'.format(
				baseline_variant))
		# Add baseline to metadata
		self.metadata['baseline_variant'] = baseline_variant

		self.dbg = dbg or Dbg()

	@property
	def baseline_variant(self):
		return self.metadata['baseline_variant']

	def __str__(self):
		res = super(Experiment, self).__str__()

		variants = self.variant_names

		res += '\n {:d} variants: {}'.format(len(variants),
											 ', '.join(
												 [('*' + k + '*') if (k == self.metadata.get('baseline_variant', '-'))
												  else k for k in variants]
											 ))
		# res += '\n KPIs are: \n   {}'.format(
		#		'\n   '.join([('**'+k+'**') if (k == self.metadata.get('primary_KPI','-')) else k for k in self.kpi_names]))

		return res

	def delta(self, kpi_subset=None, variant_subset=None,
			  assume_normal=True, percentiles=[2.5, 97.5],
			  min_observations=20, nruns=10000, relative=False):
		"""
	    Compute delta (with confidence bounds) on all applicable kpis,
	    and returns in the standard Results format.

	    Does this for all non-baseline variants.

	    TODO: Extend this function to metrics again with type-checking

	    Args:
	        kpi_subset (list): kpis for which to perfom delta. If set to
	            None all kpis are used.
	        variant_subset (list): Variants to use compare against baseline. If
	            set to None all variants are used.

	        assume_normal (boolean): specifies whether normal distribution
	            assumptions can be made
	        percentiles (list): list of percentile values to compute
	        min_observations (integer): minimum observations necessary. If
	            less observations are given, then NaN is returned
	        nruns (integer): number of bootstrap runs to perform if assume
	            normal is set to False.
	        relative (boolean): If relative==True, then the values will be
	            returned as distances below and above the mean, respectively,
	            rather than the	absolute values. In	this case, the interval is
	            mean-ret_val[0] to mean+ret_val[1]. This is more useful in many
	            situations because it corresponds with the sem() and std()
	            functions.

	    Returns:
	        Results object containing the computed deltas.
	    """
		res = Results(None, metadata=self.metadata)

		kpis_to_analyse = self.kpi_names.copy()
		if kpi_subset is not None:
			kpis_to_analyse.intersection_update(kpi_subset)
		self.dbg(3, 'kpis_to_analyse: ' + ','.join(kpis_to_analyse))

		treat_variants = self.variant_names - set([self.baseline_variant])
		self.dbg(3, 'treat_variants before subset: ' + ','.join(treat_variants))
		if variant_subset is not None:
			treat_variants.intersection_update(variant_subset)
		self.dbg(3, 'treat_variants to analyse: ' + ','.join(treat_variants))

		for mname in kpis_to_analyse:
			try:
				with warnings.catch_warnings(record=True) as w:
					# Cause all warnings to always be triggered.
					warnings.simplefilter("always")
					df = (_delta_all_variants(self.kpis.reset_index()[['entity', 'variant', mname]],
											  self.baseline_variant, assume_normal=assume_normal,
											  percentiles=percentiles, min_observations=min_observations,
											  nruns=nruns, relative=relative))
					if len(w):
						res.metadata['warnings']['Experiment.delta'] = w[-1].message

					if res.df is None:
						res.df = df
					else:
						res.df = res.df.append(df)

			except ValueError as e:
				res.metadata['errors']['Experiment.delta'] = e

		res.calculate_prob_uplift_over_zero()

		return res

	def feature_check(self, feature_subset=None, variant_subset=None,
					  threshold=0.05, percentiles=[2.5, 97.5], assume_normal=True,
					  min_observations=20, nruns=10000, relative=False):

		"""
	    Compute feature check on all features, and return dataframe with column
	    telling if feature check passed.

	    Args:
	        feature_subset (list): Features for which to perfom delta. If set to
	            None all metrics are used.
	        variant_subset (list): Variants to use compare against baseline. If
	            set to None all variants are used.
	        threshold (float): p-value used for dismissing null hypothesis (i.e.
	            no difference between features for variant and baseline).

	        assume_normal (boolean): specifies whether normal distribution
	            assumptions can be made
	        min_observations (integer): minimum observations necessary. If
	            less observations are given, then NaN is returned
	        nruns (integer): number of bootstrap runs to perform if assume
	            normal is set to False.

	    Returns:
	        pd.DataFrame containing boolean column named 'ok' stating if
	            feature chek was ok for the feature and variant combination
	            specified in the corresponding columns.
	    """
		# TODO: this should return a results structure, like all the others?
		# - can monkey patch it with a function to just get the 'ok' column

		res = Results(None, metadata=self.metadata)

		# Check if data exists TODO: Necessary or guarantted by __init__() ?
		if self.features is None:
			warnings.warn('Empty data set entered to analysis.'
						  + 'Returning empty result set')
			return res
		# TODO: Check if subsets are valid
		# If no subsets use superset
		if feature_subset is None:
			feature_subset = self.feature_names
		if variant_subset is None:
			variant_subset = self.variant_names

		# Iterate over the features
		for feature in feature_subset:
			df = (_feature_check_all_variants(self.features.reset_index()[['entity', 'variant', feature]],
											  self.baseline_variant, assume_normal=assume_normal,
											  percentiles=percentiles, min_observations=min_observations,
											  nruns=nruns, relative=relative))
			if res.df is None:
				res.df = df
			else:
				res.df = res.df.append(df)

		return res

	def sga(self, feature_subset=None, kpi_subset=None, variant_subset=None,
			n_bins=4, binning=None,
			assume_normal=True, percentiles=[2.5, 97.5],
			min_observations=20, nruns=10000, relative=False,
			**kwargs):
		"""
	    Compute subgroup delta (with confidence bounds) on all applicable
	    metrics, and returns in the standard Results format.

	    Does this for all non-baseline variants.

	    Args:
	        feature_subset (list): Features which are binned for which to
	            perfom delta computations. If set to None all features are used.
	        kpi_subset (list): KPIs for which to perfom delta computations.
	            If set to None all features are used.
	        variant_subset (list): Variants to use compare against baseline. If
	            set to None all variants are used.
	        n_bins (integer): number of bins to create if binning is None

	        binning (list of bins): preset (if None then binning is created)
	        assume_normal (boolean): specifies whether normal distribution
	            assumptions can be made
	        percentiles (list): list of percentile values to compute
	        min_observations (integer): minimum observations necessary. If
	            less observations are given, then NaN is returned
	        nruns (integer): number of bootstrap runs to perform if assume
	            normal is set to False.
	        relative (boolean): If relative==True, then the values will be
	            returned as distances below and above the mean, respectively,
	            rather than the	absolute values. In	this case, the interval is
	            mean-ret_val[0] to mean+ret_val[1]. This is more useful in many
	            situations because it corresponds with the sem() and std()
	            functions.

	    Returns:
	        Results object containing the computed deltas.
	    """
		res = Results(None, metadata=self.metadata)

		# Check if data exists
		if self.metrics is None:
			warnings.warn('Empty data set entered to analysis.'
						  + 'Returning empty result set')
			return res
		# TODO: Check if subsets are valid
		# If no subsets use superset
		if kpi_subset is None:
			kpi_subset = self.kpi_names
		if feature_subset is None:
			feature_subset = self.feature_names
		if variant_subset is None:
			variant_subset = self.variant_names
		# Remove baseline from variant_set
		variant_subset = variant_subset - set([self.baseline_variant])
		# Iterate over the kpis, features and variants
		# TODO: Check if this is the right approach,
		# groupby and unstack as an alternative?
		for kpi in kpi_subset:
			for feature in feature_subset:
				res.df = pd.concat([
					res.df,
					subgroup_deltas(
						self.metrics.reset_index()[['variant', feature, kpi]],
						variants=['dummy', self.baseline_variant],
						n_bins=n_bins,
						binning=binning,
						assume_normal=assume_normal,
						percentiles=percentiles,
						min_observations=min_observations,
						nruns=nruns, relative=relative).df])
		# Return the result object
		return res

	def trend(self, kpi_subset=None, variant_subset=None,
			  time_step=1, cumulative=True,
			  assume_normal=True, percentiles=[2.5, 97.5],
			  min_observations=20, nruns=10000, relative=False,
			  **kwargs):
		"""
	    Compute time delta (with confidence bounds) on all applicable
	    metrics, and returns in the standard Results format.

	    Does this for all non-baseline variants.

	    Args:
	        kpi_subset (list): KPIs for which to perfom delta computations.
	            If set to None all features are used.
	        variant_subset (list): Variants to use compare against baseline. If
	            set to None all variants are used.
	        time_step (integer): time increment over which to aggregate data.
	        cumulative (boolean): Trend is calculated using data from
	            start till the current bin or the current bin only

	        assume_normal (boolean): specifies whether normal distribution
	            assumptions can be made
	        percentiles (list): list of percentile values to compute
	        min_observations (integer): minimum observations necessary. If
	            less observations are given, then NaN is returned
	        nruns (integer): number of bootstrap runs to perform if assume
	            normal is set to False.
	        relative (boolean): If relative==True, then the values will be
	            returned as distances below and above the mean, respectively,
	            rather than the	absolute values. In	this case, the interval is
	            mean-ret_val[0] to mean+ret_val[1]. This is more useful in many
	            situations because it corresponds with the sem() and std()
	            functions.

	    Returns:
	        Results object containing the computed deltas.
	    """
		res = Results(None, metadata=self.metadata)
		# Check if data exists
		if self.kpis_time is None:
			warnings.warn('Empty data set entered to analysis. '
						  + 'Returning empty result set')
			res.metadata['warnings']['Experiment.trend'] = \
				UserWarning('Empty data set entered to analysis.')
			return res
		# Check if time is in dataframe column
		if 'time_since_treatment' not in self.kpis_time.index.names:
			warnings.warn('Need time column for trend analysis.'
						  + 'Returning empty result set')
			res.metadata['warnings']['Experiment.trend'] = \
				UserWarning('Need time column for trend analysis.')
			return res
		# TODO: Check if subsets are valid
		# If no subsets use superset
		if kpi_subset is None:
			kpi_subset = self.kpi_names
		if variant_subset is None:
			variant_subset = self.variant_names
		# Remove baseline from variant_set
		variant_subset = variant_subset - set([self.baseline_variant])
		# Iterate over the kpis and variants
		# TODO: Check if this is the right approach
		for kpi in kpi_subset:
			for variant in variant_subset:
				# TODO: Add metadata to res.metadata
				res_obj = time_dependent_deltas(
					self.kpis_time.reset_index()[['variant',
												  'time_since_treatment', kpi]],
					variants=[variant, self.baseline_variant],
					time_step=time_step,
					cumulative=cumulative,
					assume_normal=assume_normal,
					percentiles=percentiles,
					min_observations=min_observations,
					nruns=nruns, relative=relative)
				res.df = pd.concat([res.df, res_obj.df])

		# NB: assuming all binning objects based on the same feature are the same
		res.set_binning(res_obj.binning)
		# Return the result object
		return res


if __name__ == '__main__':
	from tests.tests_core.test_data import generate_random_data

	np.random.seed(0)
	metrics, metadata = generate_random_data()
	metrics['time_since_treatment'] = metrics['treatment_start_time']
	exp = Experiment('B', metrics, metadata, [4, 6])
	# Perform sga()
	result = exp.trend()

# result = time_dependent_deltas(data.metrics.reset_index()
#	[['variant','time_since_treatment','normal_shifted']],variants=['A','B']).df.loc[:,1]
# result = result.reset_index('subgroup',drop=True)
# result['subgroup'] = np.nan

# result = data.feature_check()
# res = _delta_all_variants(data.kpis.reset_index()[['entity','variant','normal_shifted']], 'A')
# result = data.feature_check()

# result = time_dependent_deltas(data.metrics.reset_index()
#				[['variant','treatment_start_time','normal_shifted']],variants=['A','B'])
# df = result.statistic('delta', 'uplift', 'normal_shifted', 'feature')
# result = data.delta(kpi_subset=['normal_unequal_variance'], variant_subset=['A'])



import datetime
from copy import deepcopy

import numpy as np
import pandas as pd
import version
import statistics as statx
from scipy.stats import norm

# from tests.tests_core.test_data import generate_random_data

from debugging import Dbg
from pdb import set_trace

class Results(object):
	"""
	A Results instance represents the results of a series of analyses such as
	SGA or deltaKPI.

	Q: could we make this a subclass of DataFrame (i.e. provide this class in an
	'is-a' relationship with DataFrame, rather than the 'has-a' relationship it
	has now? It seems to be a difficult thing to subclass the DataFrame object
	for some reason. https://github.com/pydata/pandas/pull/4271
	For now, will leave as a 'has-a' class.

	TODO: can we remove the 'value' level from the columns, so that the columns
	of the dataframe are simply the names of the variants? This will make the
	columns a normal index rather than a multi-index. Currently, always a multi-
	index with second level only containing a single value 'value'
	"""
	mandatory_index_levels = [
		'metric',
		'subgroup_metric',
		'subgroup',
		'statistic',
		'pctile']
	mandatory_column_levels = ['variant']

	def __init__(self, df, metadata={}, dbg=None):
		"""
	    Want to be able to create results from just a single dataframe.
	    """
		self.df = df
		self.metadata = metadata
		self.metadata['version'] = version.__version__
		self.metadata['errors'] = {}
		self.metadata['warnings'] = {}

		self.dbg = dbg or Dbg()

	@property
	def binning(self):
		"""Return the binning object."""
		return self.metadata['binning']

	def set_binning(self, binning):
		"""Store a binning object in the metadata."""
		self.metadata['binning'] = binning

	def _sortlevels(self):
		self.df.sortlevel(axis=0, inplace=True, sort_remaining=True)
		self.df.sortlevel(axis=1, inplace=True, sort_remaining=True)

	def append_delta(self, metric, variant, mu, pctiles,
					 samplesize_variant,
					 samplesize_baseline,
					 subgroup_metric='-',
					 subgroup=None):
		"""
	    Appends the results of a delta.

	    Modifies (or creates) the results data (df).
	    """
		df = delta_to_dataframe(metric, variant, mu, pctiles,
								samplesize_variant,
								samplesize_baseline,
								subgroup_metric='-',
								subgroup=None)

		if self.df is None:
			self.df = df
		else:
			self.df = self.df.append(df)

		self._sortlevels()

	def variants(self):
		"""
	    Return the variants represented in this object
	    """
		return self.df.columns.levels[0]

	def index_values(self, level='metric'):
		"""
	    Return the metrics represented in this Results object
	    """
		return self.df.index.get_level_values(level=level).unique()

	def relative_uplift(self, analysis_type, metric=None, subgroup_metric='-'):
		"""Calculate the relative uplift for the given metrics and subgroup
	    metrics.
	    """
		uplifts = self.statistic(analysis_type, 'uplift', metric, subgroup_metric)
		baseline_abs = self.statistic(analysis_type, 'variant_mean', metric,
									  subgroup_metric).loc[:, ('value', self.metadata['baseline_variant'])]
		val = uplifts.values / baseline_abs.values[:, np.newaxis]

		df = pd.DataFrame(val, columns=uplifts.columns)
		# reconstruct indices
		for i in self.mandatory_index_levels:
			df[i] = uplifts.index.get_level_values(i)
		df.set_index(self.mandatory_index_levels, inplace=True)
		# df.index.set_levels(['uplift_rel']*df.index.shape[0], level='statistic', inplace=True)

		# TODO: do we return a data frame or a Results object here?
		return df

	def calculate_prob_uplift_over_zero(self):
		# check if the subgroup index is NaN
		# NB: this will NOT work if we store delta and SGA results in the same object
		# if the subgroup index contains only NaNs
		if len(self.df.index.levels[2]) == 0:
			df = self.df.groupby(level=['metric']).apply(lambda x:prob_uplift_over_zero_single_metric(x,self.metadata['baseline_variant']))
			# remove redundant levels (coming from groupby)
			df.reset_index(level=0,drop=True,inplace=True)
		else:
			df = self.df.groupby(level=['metric','subgroup_metric','subgroup']).apply(lambda x:prob_uplift_over_zero_single_metric(x,self.metadata['baseline_variant']))
			# remove redundant levels (coming from groupby)
			df.reset_index(level=[0,1,2],drop=True,inplace=True)

		self.df = df
		#return df

	def delta_means(self, metric=None, subgroup_metric='-'):
		return self.statistic('delta', 'variant_mean', metric, subgroup_metric)

	def sga_means(self, metric=None, subgroup_metric='-'):
		return self.statistic('sga', 'variant_mean', metric, subgroup_metric)

	def uplifts(self, metric=None, subgroup_metric='-'):
		return self.statistic('delta', 'uplift', metric, subgroup_metric)

	def sga_uplifts(self, metric=None, subgroup_metric='-'):
		return self.statistic('sga', 'uplift', metric, subgroup_metric)

	def sample_sizes(self, analysis_type='delta', metric=None, subgroup_metric='-'):
		return self.statistic(analysis_type, 'sample_size', metric, subgroup_metric)

	def statistic(self, analysis_type, statistic=None, metric=None,
				  subgroup_metric='-',
				  time_since_treatment='-',
				  include_pctiles=True):
		"""

	    This is just a basic 'formatter' to allow easy access to results without
	    knowing the ordering of the index, etc. and to have sensible defaults.
	    All of this can be accomplished with fancy indexing on the dataframe
	    directly, but this should just serve as a convenience and an obvious
	    place to 'document' what the typical use-case is.

	    For all arguments, None means all, and '-' means only those for which
	    this particular argument is undefined (e.g. subgroup_metric='-')

	    Args:

	            analysis_type (string): the type of analysis that produced the  TODO: implement this!
	                results (several can be present in a single Result object). Must be
	                one of the following:
	                        - 'delta':	only those with no time_since_treatment information, and no subgroup defined
	                        - 'sga':		only those with subgroup defined
	                        - 'trend':	only those with time_since_treatment defined
	                        - None:			no restriction done

	            statistic (string): the type of data you want, such as 'uplift'

	            metric (string): which metrics you are interested in

	            time_since_treatment (int?): TODO: implement

	            include_pctiles (bool): some statistics (e.g. 'uplift') can be present
	                with percentiles defined, as opposed to just a mean. If this is true,
	                they'll be returned also. TODO: implement this!
	    """
		# NOTE: throws AssertionError for trend results and unittest example results
		# assert (self.df.index.names[0:4] == Results.mandatory_index_levels[0:4])
		# assert (self.df.columns.names[1] == Results.mandatory_column_levels[0])

		mean_results = self.df.xs((statistic, metric, subgroup_metric),
								  level=('statistic', 'metric', 'subgroup_metric'), drop_level=False)

		# metric_slicer = metric or slice(None)
		# statistic_slicer = stat or slice(None)

		# rows = (metric,subgroup_metric,slice(None),statistic_slicer)

		# levels_to_drop = ['subgroup_metric', 'statistic']
		# if subgroup_metric == '-':
		# 	levels_to_drop.append('subgroup')
		# if metric is not None:
		# 	levels_to_drop.insert(0, 'metric')

		# if stat == 'pctile':
		# 	cols = (slice(None))
		# else:
		# 	cols = (slice(None),'value')
		# 	levels_to_drop.append('pctile')

		# mean_results = self.df.loc[rows, cols]

		# #mean_results.index = mean_results.reset_index(levels_to_drop,drop=True)
		# mean_results = mean_results.reset_index(levels_to_drop,drop=True)
		# mean_results.columns = mean_results.columns.droplevel(1)


		return mean_results

	def bounds(self, metric=None, subgroup_metric='-'):

		if False:
			rows = (slice(None), '-', slice(None), ['uplift', 'uplift_pctile'])
			cols = (slice(None), 'value')
			results = self.df.loc[rows, cols].unstack(['statistic', 'pctile'])

			results.columns = results.columns.droplevel(1)

			if subgroup_metric == '-':
				results.reset_index(['subgroup', 'subgroup_metric'], drop=True)
		else:
			pctiles = self.statistic('pctile').unstack('pctile')
			mns = self.uplifts()
			mns.columns = pd.MultiIndex.from_product(
				(self.means().columns, ['mean']), names=['variant', 'statistic'])
			results = pd.concat((pctiles, mns), axis=1)
			results.columns.names = ['variant', 'statistic']

		return results

	def __str__(self):
		# TODO: improve
		return 'Results for \'{}\' with {:d} variants, {:d} metrics, {:d} subgroup metrics. Means are:\n{}'.format(
			str(self.metadata.get('experiment')),
			len(self.variants()),
			len(self.index_values('metric')),
			len(self.index_values('subgroup_metric')),
			str(self.means()),
		)

	def __repr__(self):
		return 'Results(metadata={}, \ndf={})'.format(repr(self.metadata),
													  repr(self.df.unstack('pctile')))

	def to_csv(self, fpath):
		"This will lose all metadata"
		res = deepcopy(self.df)
		res.columns = res.columns.droplevel(0)
		res = res.reset_index()
		res.to_csv(fpath, index=False)

	def to_hdf(self, fpath):
		"""
	    Persist to an HDF5 file, preserving metadata.

	    MetaData is stored as attributes on a Group called 'metadata'. This group
	    doesn't include any datasets, but was used to avoid interfering with the
	    attributes that pandas stores on the 'data' Group.

	    """
		import h5py

		store = pd.HDFStore(fpath)
		store['data'] = self.df  # write to HDF5
		store.close()

		# TODO: surely can do this via single interface rather than opening and closing?
		hfile = h5py.File(fpath)
		md = hfile.require_group('metadata')
		datetime_conversions = set(md.attrs.get('_datetime_attributes', set()))
		for k, v in self.metadata.iteritems():
			if k == '_datetime_attributes':
				continue
			if v is None:
				continue
			self.dbg(3, 'to_hdf: storing metadata {}'.format(k))
			if isinstance(v, pd.Timestamp) or isinstance(v, datetime.datetime):
				v = str(v)
				self.dbg(3, ' -> converted datetime/timestamp to string')
				datetime_conversions.add(k)
			md.attrs[k] = v

		if len(datetime_conversions) > 0:
			md.attrs['_datetime_attributes'] = [str(x) for x in datetime_conversions]

		hfile.close()


def prob_uplift_over_zero_single_metric(result_df, baseline_variant):
	"""Calculate the probability of uplift>0 for a single metric.

	Args:
		result_df (DataFrame): result data frame of a single metric/subgroup
		baseline_variant (str): name of the baseline variant

	Returns:
		DataFrame: result data frame with one additional statistic 'prob_uplift_over_0'
	"""
	pctile = 97.5 # result should be independent of the percentile that we choose
	all_variants = set(result_df.columns.levels[1])
	# iterate over all non-baseline variants
	variant = all_variants - set([baseline_variant])
	#set_trace()
	prob_dict = {baseline_variant:np.nan}
	for v in variant:
		mu = float(result_df.xs(('uplift'),level=('statistic'))[('value',v)])
		x = float(result_df.xs(('uplift_pctile',pctile),level=('statistic','pctile'))[('value',v)])
		sigma = statx.estimate_std(x, mu, pctile)

		prob = 1 - norm.cdf(0, loc=mu, scale=sigma)
		prob_dict[v] = prob

	# convert dict to df
	prob_list = []
	for v in result_df.columns.levels[1]:
		prob_list.append(prob_dict[v])
	# a prob df w/o multi-index
	prob_df = pd.DataFrame([prob_list], columns=result_df.columns)
	# reconstruct indices
	for i in ['metric','subgroup_metric','subgroup']:
		prob_df[i] = result_df.index.get_level_values(i)[0]
	prob_df['statistic'] = 'prob_uplift_over_0'
	prob_df['pctile'] = np.nan
	prob_df.set_index(Results.mandatory_index_levels, inplace=True)

	ret = pd.concat((result_df,prob_df))
	ret.sort_index(inplace=True)

	return ret

def from_hdf(fpath, dbg=None):
	"""
	Restores a Results object from HDF5 as created by the to_hdf method.
	"""
	if dbg is None:
		dbg = Dbg()

	import h5py

	data = pd.read_hdf(fpath, 'data')

	hfile = h5py.File(fpath)
	md = hfile['metadata']
	datetime_conversions = set(md.attrs.get('_datetime_attributes', set()))
	metadata = {}
	for k, v in md.attrs.iteritems():
		if k == '_datetime_attributes':
			continue
		dbg(3, 'from_hdf: retrieving metadata {}'.format(k))
		if k in datetime_conversions:
			dbg(3, ' -> converting to Timestamp')
			v = pd.Timestamp(v)
		metadata[k] = v

	return Results(data, metadata)


def delta_to_dataframe(metric, variant, mu, pctiles, samplesize_variant, samplesize_baseline,
					   subgroup_metric='-',
					   subgroup=None):
	"""Defines the Results data frame structure."""

	df = pd.DataFrame({
		'metric': metric,
		'variant': variant,
		'statistic': 'pctile',
		'pctile': pctiles.keys(),
		'value': pctiles.values(),
		'subgroup_metric': subgroup_metric,
		'subgroup': subgroup
	})
	# TODO: put baseline in as separate column... no need for sample_size_baseline
	df = df.append(pd.DataFrame({
		'metric': metric,
		'variant': variant,
		'statistic': ['mean', 'sample_size', 'sample_size_baseline'],
		'value': [mu, samplesize_variant, samplesize_baseline],
		'subgroup_metric': subgroup_metric,
		'subgroup': subgroup
	}), ignore_index=True)

	df.set_index(Results.mandatory_index_levels + ['variant'], inplace=True)
	df = df.unstack('variant')
	df.columns = df.columns.swaplevel(0, 1)

	return df


def delta_to_dataframe_all_variants(metric, mu, pctiles, samplesize_variant,
									samplesize_baseline, mu_variant,
									mu_baseline,
									subgroup_metric='-',
									subgroup=None):
	"""Defines the Results data frame structure."""

	df = pd.DataFrame({
		'metric': metric,
		'statistic': 'uplift_pctile',
		'pctile': pctiles.keys(),
		'value': pctiles.values(),
		'subgroup_metric': subgroup_metric,
		'subgroup': subgroup
	})
	df = df.append(pd.DataFrame({
		'metric': metric,
		'statistic': ['uplift', 'sample_size', 'variant_mean'],
		'value': [mu, samplesize_variant, mu_variant],
		'subgroup_metric': subgroup_metric,
		'subgroup': subgroup
	}), ignore_index=True)

	df.set_index(Results.mandatory_index_levels, inplace=True)
	# df = df.unstack('variant')
	# df.columns = df.columns.swaplevel(0,1)

	return df


def feature_check_to_dataframe(metric,
							   samplesize_variant,
							   mu=None,
							   pctiles=None,
							   pval=None,
							   mu_variant=None):
	"""Defines the Results data frame structure."""

	# numerical feature
	if pval is None:
		df = pd.DataFrame({'metric': metric,
						   'statistic': 'pre_treatment_diff_pctile',
						   'pctile': pctiles.keys(),
						   'value': pctiles.values(),
						   'subgroup_metric': '-',
						   'subgroup': None})
		df = df.append(pd.DataFrame({
			'metric': metric,
			'statistic': ['pre_treatment_diff', 'sample_size', 'variant_mean'],
			'value': [mu, samplesize_variant, mu_variant],
			'subgroup_metric': '-',
			'subgroup': None
		}), ignore_index=True)
	# categorical feature
	if mu is None:
		df = pd.DataFrame({
			'metric': metric,
			'pctile': None,
			'statistic': ['chi_square_p', 'sample_size'],
			'value': [pval, samplesize_variant],
			'subgroup_metric': '-',
			'subgroup': None
		})

	df.set_index(Results.mandatory_index_levels, inplace=True)
	# df = df.unstack('variant')
	# df.columns = df.columns.swaplevel(0,1)

	return df


if __name__ == '__main__':
	#pass
	
	np.random.seed(0)
	from tests.tests_core.test_data import generate_random_data
	from expan.core.experiment import Experiment
	data = Experiment('B', *generate_random_data())
	res = data.delta(kpi_subset=['normal_same','normal_shifted'])
	# df = res.calculate_prob_uplift_over_zero()

	# from test_core.test_results import load_example_results
	# aa = load_example_results()
	# order_means = aa.means('orders').iloc[0]
	# net_sales_var = aa.statistic('var', 'net_sales')

	# import numpy as np
	# res = Results(None)
	# res.append_delta('dummy', 'A', *(0.1,{'2.5':0.01,'97.5':0.2},1000,1000))
	# res.append_delta('dummy', 'B', *(0,{'2.5':np.nan,'97.5':np.nan},1000,1000))

	# from expan.core.experiment import Experiment
    #
	# np.random.seed(0)
	# data = Experiment('B', *generate_random_data())
	# res = data.sga()
	# x = res.relative_uplift('sga', 'normal_same', 'feature')
# res = data.delta()
# x = res.relative_uplift('delta', 'normal_same')



import unittest
import warnings

import numpy as np
import pandas as pd
from expan.core.experiment import Experiment, subgroup_deltas, time_dependent_deltas
from tests.tests_core.test_data import generate_random_data

# raise the same warning multiple times
#warnings.simplefilter('always', UserWarning)

from expan.core.debugging import Dbg


class ExperimentTestCase(unittest.TestCase):
	"""
	Defines the setUp() and tearDown() functions for the statistics test cases.
	"""

	def setUp(self):
		"""
	    Load the needed datasets for all StatisticsTestCases and set the random
	    seed so that randomized algorithms show deterministic behaviour.
	    """
		np.random.seed(0)
		self.data = Experiment('B', *generate_random_data(), dbg=Dbg(dbg_lvl=5))
		# Create time column. TODO: Do this nicer
		self.data.kpis['time_since_treatment'] = \
			self.data.features['treatment_start_time']
		# Make time part of index
		self.data.kpis.set_index('time_since_treatment', append=True, inplace=True)
		# Metadata as generated by generate_random_data() for later checks
		self.testmetadata = {'primary_KPI': 'normal_shifted',
		                     'source': 'simulated',
		                     'experiment': 'random_data_generation'}

	def tearDown(self):
		"""
	    Clean up after the test
	    """
		# TODO: find out if we have to remove data manually
		pass


class ExperimentNonClassTestCases(ExperimentTestCase):
	"""
	Test cases for the non-class functions.
	"""

	def test__subgroup_deltas__computation(self):
		"""
	    Check if subgroup_deltas() functions properly
	    """
		# Calculate result
		result = subgroup_deltas(self.data.metrics.reset_index()
								 [['variant', 'feature', 'normal_shifted']], variants=['A', 'B'])
		# check uplift
		df = result.statistic('sga', 'uplift', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-0.980599, -1.001506]), decimal=5)
		# check pctile
		df = result.statistic('sga', 'uplift_pctile', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.036771, -0.924426, -1.058535, -0.944478]), decimal=5)
		# check samplesize
		df = result.statistic('sga', 'sample_size', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[3067, 1966], [3041, 1926]]), decimal=5)
		# check variant_mean
		df = result.statistic('sga', 'variant_mean', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[0.001221, 0.981820], [-0.012310, 0.989196]]), decimal=5)


	def test__time_dependent_deltas__computation(self):
		"""
	    Check if time_dependent_deltas() functions properly
	    """
		# Calculate result
		result = time_dependent_deltas(self.data.metrics.reset_index()
									   [['variant', 'time_since_treatment', 'normal_shifted']], variants=['A', 'B'])
		# check uplift
		df = result.statistic('trend', 'uplift', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.009421, -0.847400, -1.119885, -1.042597, -0.868819,
												 -1.091165, -0.952307, -1.028234, -0.978774, -0.985696]), decimal=5)
		# check pctile
		df = result.statistic('trend', 'uplift_pctile', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.137482, -0.881360, -0.970678, -0.724122, -1.245795,
												 -0.993975, -1.178494, -0.906699, -0.993683, -0.743954, -1.225361,
												 -0.956969, -1.082180, -0.822435, -1.151715, -0.904753, -1.095209,
												 -0.862340, -1.109407, -0.861985]), decimal=5)
		# check samplesize
		df = result.statistic('trend', 'sample_size', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([649, 595, 600, 590, 625, 602, 607, 608, 616, 616],
														[405, 401, 378, 362, 377, 369, 406, 392, 414, 388])), decimal=5)
		# check variant_mean
		df = result.statistic('trend', 'variant_mean', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([0.005761, 0.057487, -0.067107, 0.001125, 0.093085,
														 -0.067894, -0.030500, -0.060996, 0.016257, -0.006091],
														[1.015182, 0.904887, 1.052778, 1.043721, 0.961904, 1.023271,
														 0.921807, 0.967238, 0.995031, 0.979605])), decimal=5)


class ExperimentClassTestCases(ExperimentTestCase):
	"""
	Test cases for the Experiment class functions.
	"""

	def test__feature_check__computation(self):
		"""
	    Check if feature check is corectly performed on test data set
	    """
		# Perform feature check
		result = self.data.feature_check()
		# check delta
		df = result.statistic('fc', 'pre_treatment_diff', 'treatment_start_time')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-0.023670]), decimal=5)
		# check pctile
		df = result.statistic('fc', 'pre_treatment_diff_pctile', 'treatment_start_time')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-0.140031, 0.092690]), decimal=5)
		# check samplesize
		df = result.statistic('fc', 'sample_size', 'treatment_start_time')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[6108, 3892]]), decimal=5)
		# check variant_mean
		df = result.statistic('fc', 'variant_mean', 'treatment_start_time')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[4.493287, 4.516958]]), decimal=5)

		# check chi-square p-values
		df = result.statistic('fc', 'chi_square_p', 'feature')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[0.769093, 1]]), decimal=5)

		# check metadata is preserved
		np.testing.assert_equal(True, all(item in result.metadata.items()
		                                for item in self.testmetadata.items()))


	def test__sga__index_levels(self):
		"""
	    Check if sga() returns the proper index levels
	    """
		# Perform sga()
		result = self.data.sga()
		# Check if all index levels are present
		index_levels = [
			pd.Index([u'normal_same', u'normal_shifted', u'normal_shifted_by_feature', u'normal_unequal_variance'],
					 dtype='object', name=u'metric'),
			pd.Index([u'feature', u'treatment_start_time'], dtype='object', name=u'subgroup_metric'),
			pd.Index([u'[0.0,2.0)', u'[2.0,4.0)', u'[4.0,7.0)', u'[7.0,9.0]', u'{has}', u'{non}'], dtype='object',
					 name=u'subgroup'),
			pd.Index([u'sample_size', u'uplift', u'uplift_pctile', u'variant_mean'], dtype='object', name=u'statistic'),
			pd.Float64Index([2.5, 97.5], dtype='float64', name=u'pctile')
		]
		result_levels = list(result.df.index.levels)
		# Check if all index levels match expectation TODO: Make nice
		np.testing.assert_array_equal(index_levels[0], result_levels[0])
		np.testing.assert_array_equal(index_levels[1], result_levels[1])
		np.testing.assert_array_equal(index_levels[2], result_levels[2])
		np.testing.assert_array_equal(index_levels[3], result_levels[3])
		np.testing.assert_array_equal(index_levels[4], result_levels[4])

	def test__sga__computation(self):
		"""
	    Check if sga() performs proper computation
	    """
		# Perform sga()
		result = self.data.sga()
		# check uplift
		df = result.statistic('sga', 'uplift', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-0.980599, -1.001506]), decimal=5)
		# check pctile
		df = result.statistic('sga', 'uplift_pctile', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.036771, -0.924426, -1.058535, -0.944478]), decimal=5)
		# check samplesize
		df = result.statistic('sga', 'sample_size', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[3067, 1966], [3041, 1926]]), decimal=5)
		# check variant_mean
		df = result.statistic('sga', 'variant_mean', 'normal_shifted', 'feature')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[0.001221, 0.981820], [-0.012310, 0.989196]]), decimal=5)

		# check metadata is preserved
		np.testing.assert_equal(True, all(item in result.metadata.items()
		                                for item in self.testmetadata.items()))


	def test__trend__index_levels(self):
		"""
	    Check if trend() returns the proper index levels
	    """
		np.random.seed(0)
		metrics, metadata = generate_random_data()
		metrics['time_since_treatment'] = metrics['treatment_start_time']
		exp = Experiment('B', metrics, metadata, [4, 6])
		# Perform sga()
		result = exp.trend()
		# Check if all index levels are present
		index_levels = [
			pd.Index([u'normal_same', u'normal_shifted', u'normal_shifted_by_feature', u'normal_unequal_variance'],
					 dtype='object', name=u'metric'),
			pd.Index([u'-'], dtype='object', name=u'subgroup_metric'),
			pd.Index([str(x) for x in np.arange(10.)], dtype='object', name=u'time'),
			pd.Float64Index([], dtype='float64', name=u'subgroup'),
			pd.Index([u'sample_size', u'uplift', u'uplift_pctile', u'variant_mean'], dtype='object', name=u'statistic'),
			pd.Float64Index([2.5, 97.5], dtype='float64', name=u'pctile')
		]
		result_levels = list(result.df.index.levels)
		# Check if all index levels match expectation TODO: Make nice
		np.testing.assert_array_equal(index_levels[0], result_levels[0])
		np.testing.assert_array_equal(index_levels[1], result_levels[1])
		np.testing.assert_array_equal(index_levels[2], result_levels[2])
		np.testing.assert_array_equal(index_levels[3], result_levels[3])
		np.testing.assert_array_equal(index_levels[4], result_levels[4])
		np.testing.assert_array_equal(index_levels[5], result_levels[5])

	def test__trend__computation(self):
		"""
	    Check if trend() functions properly
	    """
		np.random.seed(0)
		metrics, metadata = generate_random_data()
		metrics['time_since_treatment'] = metrics['treatment_start_time']
		exp = Experiment('B', metrics, metadata, [4, 6])
		# Perform sga() with non-cumulative results
		result = exp.trend(cumulative=False)

		# check uplift
		df = result.statistic('trend', 'uplift', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.009421, -0.847400, -1.119885, -1.042597, -0.868819,
												 -1.091165, -0.952307, -1.028234, -0.978774, -0.985696]), decimal=5)
		# check pctile
		df = result.statistic('trend', 'uplift_pctile', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.137482, -0.881360, -0.970678, -0.724122, -1.245795,
												 -0.993975, -1.178494, -0.906699, -0.993683, -0.743954, -1.225361,
												 -0.956969, -1.082180, -0.822435, -1.151715, -0.904753, -1.095209,
												 -0.862340, -1.109407, -0.861985]), decimal=5)
		# check samplesize
		df = result.statistic('trend', 'sample_size', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([649, 595, 600, 590, 625, 602, 607, 608, 616, 616],
														[405, 401, 378, 362, 377, 369, 406, 392, 414, 388])), decimal=5)
		# check variant_mean
		df = result.statistic('trend', 'variant_mean', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([0.005761, 0.057487, -0.067107, 0.001125, 0.093085,
														 -0.067894, -0.030500, -0.060996, 0.016257, -0.006091],
														[1.015182, 0.904887, 1.052778, 1.043721, 0.961904, 1.023271,
														 0.921807, 0.967238, 0.995031, 0.979605])), decimal=5)

		# Perform sga() with cumulative results
		result = exp.trend()
		# check uplift
		df = result.statistic('trend', 'uplift', 'normal_shifted')

		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-1.009421, -0.929807, -0.991088, -1.003129, -0.976023,
												 -0.994857,-0.988167,-0.993119,-0.991571, -0.990986]), decimal=5)
		# check pctile
		df = result.statistic('trend', 'uplift_pctile', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([ -1.137482, -0.881360, -1.018794, -0.840820, -1.063820,
									              -0.918356, -1.067283, -0.938976, -1.033110, -0.918936,
									              -1.047413, -0.942302, -1.036888, -0.939446, -1.038455,
									              -0.947784, -1.033861, -0.949280, -1.031002, -0.950970]), decimal=5)
		# check samplesize
		df = result.statistic('trend', 'sample_size', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([ 649, 1244, 1844, 2434, 3059, 3661, 4268, 4876, 5492, 6108],
														[  405, 806, 1184, 1546, 1923, 2292, 2698, 3090, 3504, 3892])), decimal=5)
		# check variant_mean
		df = result.statistic('trend', 'variant_mean', 'normal_shifted')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.column_stack(([ 0.005761, 0.030501, -0.001258, -0.000681, 0.018477,
									                      0.004274, -0.000671, -0.008193, -0.005451, -0.005515],
														[ 1.015182, 0.960308, 0.989830, 1.002449, 0.994500,
														  0.999132, 0.987496, 0.984926, 0.986120, 0.985470])), decimal=5)

		# check metadata is preserved
		np.testing.assert_equal(True, all(item in result.metadata.items()
		                                for item in self.testmetadata.items()))


	def test_trend_missing_time_resolved_data(self):
		"""Check if missing time-resolved data in trend analysis issues a warning"""
		result = self.data.trend()
		w = result.metadata['warnings']['Experiment.trend']
		self.assertTrue(isinstance(w, UserWarning))
		self.assertTrue(w.message == 'Empty data set entered to analysis.')



	def test_delta(self):
		"""
	    Check if Experiment.delta() functions properly
	    """
		# this should work
		self.assertTrue(isinstance(self.data, Experiment))  # check that the subclassing works

		self.assertTrue(self.data.baseline_variant == 'B')

		result = self.data.delta(kpi_subset=
								 [m for m in self.data.kpi_names if 'normal' in m])

		# check uplift
		df = result.statistic('delta', 'uplift', 'normal_same')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([0.033053]), decimal=5)
		# check pctile
		df = result.statistic('delta', 'uplift_pctile', 'normal_same')
		np.testing.assert_almost_equal(df.loc[:, ('value', 'A')],
									   np.array([-0.007135, 0.073240]), decimal=5)
		# check samplesize
		df = result.statistic('delta', 'sample_size', 'normal_same')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[6108, 3892]]), decimal=5)
		# check variant_mean
		df = result.statistic('delta', 'variant_mean', 'normal_same')
		np.testing.assert_almost_equal(df.loc[:, 'value'],
									   np.array([[0.025219, -0.007833]]), decimal=5)

		# check metadata is preserved
		np.testing.assert_equal(True, all(item in result.metadata.items()
		                                for item in self.testmetadata.items()))

	def test_unequal_variance_warning_in_results(self):
		"""
		Check if the unequal variance warning message is persisted to the Results structure
    	"""
		result = self.data.delta(kpi_subset=['normal_unequal_variance'],
								 variant_subset=['A'])
		w = result.metadata['warnings']['Experiment.delta']
		self.assertTrue(isinstance(w, UserWarning))
		self.assertTrue(w.message == 'Sample variances differ too much to assume that population variances are equal.')


if __name__ == '__main__':
	unittest.main()



import unittest
import warnings

import pandas as pd
# import inspect
import os
import numpy as np
import expan.core.statistics as statx

# import random

reload(statx)

data_dir = os.getcwd() + '/tests/tests_core/'  # TODO: adjust this depending on where we're calling


# TODO: include this functions in some more general module
def get_norm_temp_data(fname='normtemp.dat.txt.gz'):
	"""
  	Data retrieved on 2015/02/18 from:
      http://www.amstat.org/publications/jse/jse_data_archive.htm
  	"""

	# Read data from csv to pd.dataFrame
	data = pd.read_csv(
		os.path.join(data_dir, fname),
		delim_whitespace=True,
		header=None,
		skip_blank_lines=True,
		names=['temperature', 'gender', 'heartrate'],
	)

	# Return the pd.dataFrame
	return data


# TODO: include this functions in some more general module
def get_framingham_data(fname='framingham_heart_study_exam7.csv'):
	"""
      Data retrieved on 2015/10/28 from:
          http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html
  	"""

	# Read data from csv to pd.dataFrame
	data = pd.read_csv(os.path.join(data_dir, fname),
					   header=[0, 1],
					   index_col=0)
	# Set index
	data.index.name = 'metric'

	# Read data from csv to pd.dataFrame
	return data


class StatisticsTestCase(unittest.TestCase):
	"""
  	Defines the setUp() and tearDown() functions for the statistics test cases.
  	"""

	def setUp(self):
		"""
    	Load the needed datasets for all StatisticsTestCases and set the random
    	seed so that randomized algorithms show deterministic behaviour.
    	"""
		np.random.seed(0)
		self.samples = get_norm_temp_data()
		self.metrics = get_framingham_data()

		self.rand_s1 = np.random.normal(loc=0, size=1000)
		self.rand_s2 = np.random.normal(loc=0.1, size=1000)

	def tearDown(self):
		"""
    	Clean up after the test
    	"""
		# TODO: find out if we have to remove data manually
		pass

	# def test_estimate_std(self):
	# 	"""
	# 	Check if the std estimation is correct.

	# 	http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html
	# 	"""
	# 	res = statx.estimate_std(0.44, 1.7, 2.5, 1623, 1911)
	# 	self.assertAlmostEqual(res, 19.03837750)
	# 	res = statx.estimate_std(2.96, 1.7, 97.5, 1623, 1911)
	# 	self.assertAlmostEqual(res, 19.03837750)

class DeltaTestCases(StatisticsTestCase):
	"""
  	Test cases for the delta() function in core.statistics.
  	"""

	def test__delta__not_providing_data_fails(self):
		"""
    	Value error raised when not providing data.
    	"""
		# Check if error is raised for None data
		with self.assertRaises(ValueError):
			statx.delta(self.samples.temperature, None)
		# Check if error is raised for None data
		with self.assertRaises(ValueError):
			statx.delta(None, self.samples.temperature)

	def test__delta__computation_assumed_normal(self):
		"""
    	Result of delta() assuming normality equals expected result.
    	"""
		# Computing delta assumed normal
		result1 = statx.delta(
			self.samples.temperature[self.samples.gender == 1],
			self.samples.temperature[self.samples.gender == 2],
			percentiles=[2.5, 97.5],
			assume_normal=True)
		# Checking if mean has right value
		self.assertAlmostEqual(result1[0], -0.28923076923075541)
		# Checking if lower percentile has right value
		self.assertAlmostEqual(result1[1][2.5], -0.53770569567692295)
		# Checking if uper percentile has right value
		self.assertAlmostEqual(result1[1][97.5], -0.040755842784587965)
		# Checking if sample size 1 is correct
		self.assertEqual(result1[2], 65)
		# Checking if sample size 2 is correct
		self.assertEqual(result1[3], 65)

	def test__delta__nan_handling(self):
		"""
    	Test correct handling of nans. (ignored)
    	"""
		result1 = statx.delta(self.rand_s1, self.rand_s2)
		self.assertEqual(result1[2], 1000)
		self.assertEqual(result1[3], 1000)

		r1 = self.rand_s1.copy();
		r1[90:] = np.nan
		(uplift, pctiles, ss_x, ss_y, mean_x, mean_y) = statx.delta(r1, self.rand_s2)

		self.assertAlmostEqual(uplift, -0.1, 1)
		self.assertEqual(ss_x, 90)
		self.assertEqual(ss_y, 1000)

	def test__delta__computation_not_assumed_normal(self):
		"""
    	Result of delta() not assuming normality equals expected result.
    	"""
		# Computing delta not assumed normal
		result1 = statx.delta(
			self.samples.temperature[self.samples.gender == 1],
			self.samples.temperature[self.samples.gender == 2],
			percentiles=[2.5, 97.5],
			assume_normal=True)
		# Checking if mean has right value
		self.assertAlmostEqual(result1[0], -0.28923076923075541)
		# Checking if lower percentile has right value
		self.assertAlmostEqual(result1[1][2.5], -0.53770569567692295)
		# Checking if uper percentile has right value
		self.assertAlmostEqual(result1[1][97.5], -0.040755842784587965)
		# Checking if sample size 1 is correct
		self.assertEqual(result1[2], 65)
		# Checking if sample size 2 is correct
		self.assertEqual(result1[3], 65)


class ChiSquareTestCases(StatisticsTestCase):
	"""
  	Test cases for the chi_square() function in core.statistics.
  	"""

	def test__chi_square__not_providing_data_fails(self):
		"""
    	Value error raised when not providing data.
    	"""
		# Check if error is raised for None data
		with self.assertRaises(ValueError):
			statx.chi_square(self.samples.temperature, None)
		# Check if error is raised for None data
		with self.assertRaises(ValueError):
			statx.chi_square(None, self.samples.temperature)

	def test__chi_square__computation_same_data(self):
		"""
    	Check if p-value is 1.0 for same data entered twice.
    	"""
		# Computation of chi-square p-value
		self.assertEqual(1.0,
						 statx.chi_square(self.samples.temperature,
										  self.samples.temperature)[0])

	def test__chi_square__computation_different_data(self):
		"""
    	Check if p-value is correct for test data from pandas manual page.
    	"""
		# Create test data:
		a = ['A'] * 16 + ['B'] * 18 + ['C'] * 16 + ['D'] * 14 + ['E'] * 12 + ['F'] * 12
		b = ['A'] * 16 + ['B'] * 16 + ['C'] * 16 + ['D'] * 16 + ['E'] * 16 + ['F'] * 8
		# Computation of chi-square p-value
		self.assertAlmostEqual(0.89852623940266074, statx.chi_square(a, b)[0])

	def test__chi_square__computation_different_data_as_in_statistics_book(self):
		"""
    	Check if p-value is correct for test data from statistics book
    	Fahrmeir et al. (2007) pp. 463.
    	"""
		# Create test data:
		a = ['nein'] * 139 + ['gut'] * 348 + ['mittel'] * 213
		b = ['nein'] * 135 + ['gut'] * 46 + ['mittel'] * 119
		# Computation of chi-square p-value
		p, chisq, nattr = statx.chi_square(a, b)
		self.assertAlmostEqual(116.851, chisq, delta=0.001)
		self.assertAlmostEqual(0.0, p, delta=0.00000000001)

	def test__chi_square__computation_different_data_as_in_open_statistics_book(self):
		"""
    	Check if p-value is correct for test data from
    	open statistics book 3rd ed pp. 299.
    	(https://www.openintro.org/stat/textbook.php)
    	"""
		# Create test data:
		a = ['cu'] * 3511 + ['t1'] * 1749 + ['t2'] * 1818
		b = ['cu'] * 1489 + ['t1'] * 751 + ['t2'] * 682
		# Computation of chi-square p-value
		p, chisq, nattr = statx.chi_square(a, b)
		self.assertAlmostEqual(6.120, chisq, delta=0.001)
		self.assertAlmostEqual(0.0469, p, delta=0.0001)

	def test__chi_square__computation_one_bin_not_present(self):
		"""
    	Check if p-value is correct for test data from pandas manual page.
    	"""
		# Create test data:
		a = ['A'] * 16 + ['B'] * 18 + ['C'] * 16 + ['D'] * 14 + ['E'] * 12 + ['F'] * 12
		b = ['A'] * 16 + ['B'] * 16 + ['C'] * 16 + ['D'] * 16 + ['E'] * 16 + ['F'] * 8
		# Computation of chi-square p-value (a is shortened)
		self.assertAlmostEqual(0.94879980715092971, statx.chi_square(a[0:-12], b)[0])
		# Computation of chi-square p-value (b is shortened)
		self.assertAlmostEqual(0.94879980715092971, statx.chi_square(a, b[0:-8])[0])

	def test__chi_square__computation_symmetric(self):
		"""
    	Check if p-value is roughly symmetric.
    	"""
		# Create test data:
		a = ['A'] * 16 + ['B'] * 18 + ['C'] * 16 + ['D'] * 14 + ['E'] * 12 + ['F'] * 12
		b = ['A'] * 16 + ['B'] * 16 + ['C'] * 16 + ['D'] * 16 + ['E'] * 16 + ['F'] * 8
		# Computation of chi-square p-value (a is shortened)
		self.assertAlmostEqual(statx.chi_square(a, b), statx.chi_square(b, a))
		# Computation of chi-square p-value (b is shortened)
		aa = statx.chi_square(b[0:(-8)], a)
		bb = statx.chi_square(a, b[0:(-8)])
		self.assertAlmostEqual(aa[0], bb[0])  # p-value
		self.assertAlmostEqual(aa[1], bb[1])  # chi-square value


class SampleSizeTestCases(StatisticsTestCase):
	"""
  	Test cases for the sample_size() function in core.statistics.
  	"""

	def test__sample_size__empty_list_numeric(self):
		"""
    	Empty list returns 0.

    	"""
		self.assertEqual(statx.sample_size([]), 0)

	def test__sample_size__list_numeric(self):
		"""
    	Result of sample_size() is number of elements of a list.
    	"""
		x = [1, 1, 2, 5, 8]
		self.assertEqual(statx.sample_size(x), 5)

	def test__sample_size__nparray_numeric_with_nan(self):
		"""
    	Result of sample_size() is number of elements of numpy array minus
    	number of NaNs.
    	"""
		x = np.array([1, 1, np.nan, 2, np.nan, 5, 8])
		self.assertEqual(statx.sample_size(x), 5)

	def test__sample_size__list_categorical(self):
		"""
    	Result of sample_size() is number of elements of a list of categorical
    	data.
    	"""
		x = ['1', '1', '3', '2', '6', '5', '8']
		self.assertEqual(statx.sample_size(x), 7)

	def test__sample_size__pdseries_categorical(self):
		"""
    	Result of sample_size() is number of elements of a pandas series with
    	categorical data.
    	"""
		x = pd.Series(['1', '7', '2', '5', '8', '0'])
		self.assertEqual(statx.sample_size(x), 6)

	def test__sample_size__pdseries_categorical_with_na(self):
		"""
    	Result of sample_size() is number of elements of a pandas series with
    	categorical data including NAs.
    	"""
		x = ['1', '1', 'NA', '2', 'NA', '5', '8']
		self.assertEqual(statx.sample_size(x), 5)


class AlphaToPercentilesTestCases(StatisticsTestCase):
	"""
  	Test cases for the alpha_to_percentiles() function in core.statistics.
  	"""

	def test__alpha_to_percentiles__computation(self):
		"""
    	Result of alpha_to_percentiles() equals expected result.
    	"""
		# Computing alpha_to_percentiles
		result = statx.alpha_to_percentiles(0.05)
		# Checking if first percentile is correct
		self.assertEqual(result[0], 2.5)
		# Checking if second percentile is correct
		self.assertEqual(result[1], 97.5)


class BootstrapTestCases(StatisticsTestCase):
	"""
  	Test cases for the bootstrap() function in core.statistics.
  	"""

	def setUp(self):
		"""
    	Do the same setup as before - separate so that changing things in the base
    	TestCase doesn't change the randomisation here.
    	"""

		np.random.seed(0)
		self.samples = get_norm_temp_data()
		self.metrics = get_framingham_data()

	def test__bootstrap__computation(self):
		"""
    	Result of bootstrap() equals expected result.
    	"""
		# Computing bootstrap
		result1 = statx.bootstrap(self.samples.temperature, [0],
								  min_observations=1)
		# Checking if lower percentile of result2 is correct
		self.assertAlmostEqual(result1[0][2.5], 98.1220, 2)
		# Checking if upper percentile of result2 is correct
		self.assertAlmostEqual(result1[0][97.5], 98.3765, 2)
		# Checking if no bootstrap data was passed
		self.assertIsNone(result1[1])

		# Defining data and computing bootstrap
		zero_3 = np.array([0., 0., 0.])
		one_3 = np.array([1., 1., 1.])
		result2 = statx.bootstrap(zero_3, one_3, min_observations=3)
		# Checking if lower percentile of result2 is correct
		self.assertEqual(result2[0][2.5], -1.0)
		# Checking if upper percentile of result2 is correct
		self.assertEqual(result2[0][97.5], -1.0)
		# Checking if no bootstrap data was passed
		self.assertIsNone(result2[1])

		# Defining data and computing bootstrap
		sample1 = self.samples.temperature[self.samples.gender == 1]
		sample2 = self.samples.temperature[self.samples.gender == 2]
		result3 = statx.bootstrap(sample1, sample2)
		# Checking if lower percentile of result3 is correct
		self.assertAlmostEqual(result3[0][2.5], -0.53384615384615619)
		# Checking if upper percentile of result3 is correct
		self.assertAlmostEqual(result3[0][97.5], -0.041538461538493367)
		# Checking if no bootstrap data was passed
		self.assertIsNone(result3[1])


class PooledStdTestCases(StatisticsTestCase):
	"""
  	Test cases for the pooled_std() function in core.statistics.
  	"""

	def test__pooled_std__variances_differ_too_much_error(self):
		"""
    	Warning raised when variances differ too much.
    	"""
		with warnings.catch_warnings(record=True) as w:
			warnings.simplefilter('always')

			statx.pooled_std(0.25, 4, 0.5, 4)

			self.assertEqual(len(w), 1)
			self.assertTrue(issubclass(w[-1].category, UserWarning))
			self.assertTrue('variances differ' in str(w[-1].message))

	def test__pooled_std__computation(self):
		"""
    	Result of pooled_std() equals expected result.
    	"""
		# Define subset of data for test
		sbp = self.metrics.loc['systolic_bp', :]
		# Computation of pooled standard deviation
		result1 = statx.pooled_std(sbp.men.s, sbp.men.n,
								   sbp.women.s, sbp.women.n)
		# Checking if result1 is correct
		self.assertAlmostEqual(result1, 19, 1)

		# Computation of pooled standard deviation
		result2 = statx.pooled_std(9.7, 6, 12.0, 4)
		# Checking pooled_std() on small sample subset
		self.assertAlmostEqual(result2, 10.6, 1)


class NormalSamplePercentilesTestCases(StatisticsTestCase):
	"""
  	Test cases for the normal_sample_percentiles() function in core.statistics.
  	"""

	def test__normal_sample_percentiles__computation(self):
		"""
    	Result of normal_sample_percentiles() equals expected result.

    	Example from:
        http://www.stat.yale.edu/Courses/1997-98/101/confint.htm
    	"""
		# Define data for testing
		val = np.array([102.5, 101.7, 103.1, 100.9, 100.5, 102.2])
		# Computation with relative set True
		result1 = statx.normal_sample_percentiles(val, relative=False)
		# Checking if lower percentile of result1 is correct
		self.assertAlmostEqual(result1[2.5], 100.87330892117532)
		# Checking if upper percentile of result1 is correct
		self.assertAlmostEqual(result1[97.5], 102.760024412158)

		# Computation with relative set True
		result2 = statx.normal_sample_percentiles(val, relative=True)
		# Checking if lower percentile of result2 is correct
		self.assertAlmostEqual(result2[2.5], -0.94335774549134732)
		# Checking if upper percentile of result2 is correct
		self.assertAlmostEqual(result2[97.5], 0.94335774549134732)


class NormalDifferenceTestCases(StatisticsTestCase):
	"""
  	Test cases for the normal_difference() function in core.statistics.
  	"""

	def test__normal_difference__computation(self):
		"""
    	Result of normal_difference() equals expected result.
    	"""
		# Define subset of data for first test
		sbp = self.metrics.loc['systolic_bp', :]
		# computation of normal difference
		result1 = statx.normal_difference(sbp.men.m, sbp.men.s, sbp.men.n,
										  sbp.women.m, sbp.women.s, sbp.women.n)
		# Checking if lower percentile of result1 is correct
		self.assertAlmostEqual(result1[2.5], 0.44582598543756413)
		# Checking if upper percentile of result1 is correct
		self.assertAlmostEqual(result1[97.5], 2.9541740145624127)

		# Define subset of data for second test
		clst = self.metrics.loc['serum_cholesterol', :]
		# Computation of normal difference
		result2 = statx.normal_difference(clst.men.m, clst.men.s, clst.men.n,
										  clst.women.m, clst.women.s, clst.women.n)
		# Checking if lower percentile of result2 is correct
		self.assertAlmostEqual(result2[2.5], -17.159814380797162)
		# Checking if upper percentile of result2 is correct
		self.assertAlmostEqual(result2[97.5], -12.240185619202816)

		# test subsample of systolic blood pressure. Example from:
		# http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html
		# Computation of normal difference
		result3 = statx.normal_difference(117.5, 9.7, 6, 126.8, 12., 4)
		# Checking if lower percentile of result3 is correct
		self.assertAlmostEqual(result3[2.5], -25.10960582643531)
		# Checking if upper percentile of result3 is correct
		self.assertAlmostEqual(result3[97.5], 6.5096058264353118)


class NormalSampleDifferenceTestCases(StatisticsTestCase):
	"""
  	Test cases for the normal_sample_difference() function in core.statistics.
  	"""

	def test__normal_sample_difference__computation(self):
		# Defining data and computing normal difference
		sample1 = self.samples.temperature[self.samples.gender == 1]
		sample2 = self.samples.temperature[self.samples.gender == 2]
		# Computation of normal sample difference
		result1 = statx.normal_sample_difference(sample1, sample2)
		# Checking if lower percentile of result1 is correct
		self.assertAlmostEqual(result1[2.5], -0.53770569567692295)
		# Checking if upper percentile of result1 is correct
		self.assertAlmostEqual(result1[97.5], -0.040755842784587965)


if __name__ == '__main__':
	unittest.main()



'''
This started as a very simple central place to provide a really easy interface
to output (standardised across projects) so I could always just do dbg(1, 'blah').

It has morphed and mutated into an ugly wrapper on top of logger.

I keep it because I still find logger to be a PITA and too complicated to use
quickly.

A few things it offers over just using logging:
	- defaults to DEBUG level
	- use positive and negative numbers to configure level
	- standardised logging msg
	- easier calling `dbg(1, 'blah')` - dubious advantage

Created on Aug 11, 2014

TODO: need to figure out how to tell the logging class to skip this module's functions
in the stack, for printing of module and function name etc.. Could potentially
fiddle with the stack before calling logging, but that's very dodgy.

@author: rmuil
'''
import inspect
import logging
import os

default_dbg_lvl = 2

_stackframe_filename_idx = 1
_stackframe_lineno_idx = 2
_stackframe_funcname_idx = 3


def _wherefrom(skip_stack_lvls=0):
	"""
	Returns the name of the function from which this was called.

	NB: from https://docs.python.org/2/library/inspect.html:
	"Keeping references to frame objects, as found in the first element of the
	frame records these functions return, can cause your program to create
	reference cycles..."

	In the stack, pos 0 is this function, 1 is direct caller of this function
	and 2 is the caller of the caller, which would usually be the interesting one.

	NB: this seems to throw an IndexError at the inspect.stack() line sometimes. HAven't traced the bug yet.
	"""
	stack_lvl = 1 + skip_stack_lvls
	caller_name = caller_file = None
	caller_lineno = -1
	current_stack = inspect.stack()
	try:
		caller_name = current_stack[stack_lvl][_stackframe_funcname_idx]
		caller_file = os.path.basename(current_stack[stack_lvl][_stackframe_filename_idx])
		caller_lineno = current_stack[stack_lvl][_stackframe_lineno_idx]
	finally:
		del current_stack

	if caller_file.startswith('<ipython-input'):
		caller_file = '<ipython>'
		caller_name = '<prompt>'

	return (caller_name, caller_file, caller_lineno)


class Dbg():
	def __init__(self, logger=None, dbg_lvl=default_dbg_lvl):
		if logger is None:
			logging.basicConfig(level=logging.DEBUG)  # configure the root logger
			logger = logging.getLogger()  # get the root logger
		self.logger = logger
		self._dbg_lvl = dbg_lvl
		self._call_count = {}

	def __call__(self, lvl, msg, skip_stack_lvls=0):
		return self.out(lvl, msg, skip_stack_lvls=skip_stack_lvls + 1)

	def out(self, lvl, msg,
			skip_stack_lvls=0):
		"""
		This isn't very well implemented for important messages...
		we should move to relying much more on the Logging class for dispatching
		and filtering logs.

		Treat this as just a convenience.
		"""
		caller_name, caller_file, caller_lineno = _wherefrom(skip_stack_lvls + 1)
		dlvl = self._dbg_lvl
		if caller_file not in self._call_count:
			self._call_count[caller_file] = 1
		else:
			self._call_count[caller_file] += 1
		if lvl <= dlvl:
			if lvl == 0:
				self.logger.info('%s,%d|%s: %s' % (caller_file, caller_lineno, caller_name, msg))
			else:
				if lvl == -1:
					self.logger.warning('%s,%d|%s: %s' % (caller_file, caller_lineno, caller_name, msg))
				elif lvl < -1:
					self.logger.error('%s,%d|%s: %s' % (caller_file, caller_lineno, caller_name, msg))
				else:
					self.logger.debug('D%d|%s,%d|%s: %s' % (lvl, caller_file, caller_lineno, caller_name, msg))

	def set_lvl(self, new_dbg_lvl):
		self._dbg_lvl = new_dbg_lvl

	def get_lvl(self):
		return self._dbg_lvl

	def reset(self):
		self._dbg_lvl = default_dbg_lvl
		self._call_count = {}



# import os
import unittest

import numpy as np

import expan.core.util as util


class UtilFunctionsTestCase(unittest.TestCase):
	def test_scale_range(self):
		res = util.scale_range([1, 3, 5])
		np.testing.assert_allclose(res, [0., 0.5, 1.])

		res = util.scale_range([1, 2, 3, 4, 5])
		np.testing.assert_allclose(res, [0., 0.25, 0.5, 0.75, 1.])

		res = util.scale_range([1, 3, 5, np.inf])
		np.testing.assert_allclose(res, [0., 0.5, 1., np.inf])

		res = util.scale_range([1, 3, 5, -np.inf])
		np.testing.assert_allclose(res, [0., 0.5, 1., -np.inf])

		res = util.scale_range([1, 3, 5, -np.inf], squash_inf=True)
		np.testing.assert_allclose(res, [0., 0.5, 1., 0.])

		res = util.scale_range([1, 3, 5, np.inf], squash_inf=True)
		np.testing.assert_allclose(res, [0., 0.5, 1., 1.])

		res = util.scale_range([1, 3, 5], new_min=0.5)
		np.testing.assert_allclose(res, [0.5, 0.75, 1.])

		res = util.scale_range([1, 3, 5], old_min=1, old_max=4)
		np.testing.assert_allclose(res, [0., 0.66666667, 1.])

		res = util.scale_range([5], old_max=4)
		np.testing.assert_allclose(res, [1.])


if __name__ == '__main__':
	unittest.main()



import unittest
import warnings
from copy import deepcopy

import numpy as np
import pandas as pd

from expan.core.experimentdata import ExperimentData


def generate_random_data():
	np.random.seed(42)
	size = 10000

	test_data_frame = pd.DataFrame()
	test_data_frame['entity'] = range(size)
	test_data_frame['variant'] = np.random.choice(['A', 'B'], size=size, p=[0.6, 0.4])

	test_data_frame['normal_same'] = np.random.normal(size=size)
	test_data_frame['normal_shifted'] = np.random.normal(size=size)

	test_data_frame.loc[test_data_frame['variant'] == 'B', 'normal_shifted'] \
		= np.random.normal(loc=1.0, size=test_data_frame['normal_shifted'][test_data_frame['variant'] == 'B'].shape[0])

	test_data_frame['feature'] = np.random.choice(['has', 'non'], size=size)

	test_data_frame['normal_shifted_by_feature'] = np.random.normal(size=size)

	randdata = np.random.normal(loc=1.0, size=size)
	ii = (test_data_frame['variant'] == 'B') & (test_data_frame['feature'] == 'has')

	with warnings.catch_warnings(record=True) as w:
		# ignore the 'flat[index' warning that comes out of pandas (and is
		# not ours to fix)
		warnings.simplefilter('ignore', DeprecationWarning)

		test_data_frame.loc[ii, 'normal_shifted_by_feature'] = randdata

	test_data_frame['treatment_start_time'] = np.random.choice(range(10), size=size)

	test_data_frame['normal_unequal_variance'] = np.random.normal(size=size)
	test_data_frame.loc[test_data_frame['variant'] == 'B', 'normal_unequal_variance'] \
		= np.random.normal(scale=10,
						   size=test_data_frame['normal_unequal_variance'][test_data_frame['variant'] == 'B'].shape[0])

	metadata = {
		'primary_KPI': 'normal_shifted',
		'source': 'simulated',
		'experiment': 'random_data_generation'
	}

	return test_data_frame, metadata


def generate_random_data_n_variants(n_variants=3):
	np.random.seed(42)
	size = 10000

	test_data_frame = pd.DataFrame()
	test_data_frame['entity'] = range(size)
	test_data_frame['variant'] = np.random.choice(map(chr, range(65,65+n_variants)), size=size)

	test_data_frame['normal_same'] = np.random.normal(size=size)
	test_data_frame['poisson_same'] = np.random.poisson(size=size)

	test_data_frame['feature'] = np.random.choice(['has', 'non'], size=size)

	test_data_frame['treatment_start_time'] = np.random.choice(range(10), size=size)

	metadata = {
		'primary_KPI': 'normal_same',
		'source': 'simulated',
		'experiment': 'random_data_generation'
	}

	return test_data_frame, metadata


class DataTestCase(unittest.TestCase):
	def setUp(self):
		"""
    Load the needed datasets for all StatisticsTestCases and set the random
    seed so that randomized algorithms show deterministic behaviour.
    """
		# np.random.seed(0)
		self.metrics, self.metadata = generate_random_data()

	def tearDown(self):
		"""
    Clean up after the test
    """
		# TODO: find out if we have to remove data manually
		pass

	def test_create_with_insufficient_data(self):
		# should not work:
		with self.assertRaises(KeyError):
			ExperimentData(
				pd.DataFrame(columns=['entity', 'variant']),
				metadata={'experiment': 'test', 'source': 'none'}
			)
		with self.assertRaises(KeyError):
			ExperimentData(
				pd.DataFrame(columns=['entity', 'variant', 'plums']),
				metadata={'experiment': 'test', 'source': 'none', 'primary_KPI': 'plums'}
			)
		with self.assertRaises(ValueError):
			ExperimentData(None)
		# with self.assertRaises(KeyError):
		#	ExperimentData(pd.DataFrame())
		with self.assertRaises(KeyError):
			ExperimentData(
				pd.DataFrame(columns=['entity', 'variant']),
				metadata=None)
		# with self.assertRaises(KeyError):
		# 	ExperimentData(
		# 		pd.DataFrame(columns=['entity', 'treatment_start']),
		# 		metadata={'experiment': 'fesf', 'source': 'random'},
		# 	)
		with self.assertRaises(KeyError):
			ExperimentData(
				pd.DataFrame(columns=['variant', 'treatment_start']),
				metadata={'experiment': 'fesf', 'source': 'random'}
			)
		with self.assertRaises(KeyError):
			ExperimentData(
				pd.DataFrame(columns=['variant', 'entity', 'treatment_start']),
				metadata={
					'experiment': 'fesf',
					'source': 'random',
					'primary_KPI': 'something_not_there'}
			)

	def test_data_generation(self):
		df, md = generate_random_data()
		A = ExperimentData(df, md)
		self.assertIsNotNone(A.kpis)
		self.assertIsNotNone(A.features)
		self.assertIsNotNone(A.metadata)

		# also test incomplete info again
		with self.assertRaises(KeyError):
			ExperimentData(df.drop('entity', axis=1), md)
		# with self.assertRaises(KeyError):
		# 	ExperimentData(df.drop('variant', axis=1), md)

	def test_direct_indexing(self):
		A = ExperimentData(*generate_random_data())
		# this should work
		normal_shifted = A[['normal_shifted', 'feature']]

		# this should not
		with self.assertRaises(KeyError):
			normal_shifted = A[['normal_shifted', 'non_existent_feature']]

	def test_initialize_without_kpi_with_feature(self):
		"""Initialize ExperimentData with metrics=None, features=DF"""
		metrics, _ = generate_random_data()
		meta = {
			'source': 'simulated',
			'experiment': 'random_data_generation'
		}
		D = ExperimentData(None, meta, metrics)

	def test_initialize_without_kpi_without_feature(self):
		"""Initialize ExperimentData with metrics=None, features=[]/'default'"""
		meta = {
			'source': 'simulated',
			'experiment': 'random_data_generation'
		}

		with self.assertRaises(ValueError):
			D = ExperimentData(None, meta, [])
		with self.assertRaises(ValueError):
			D = ExperimentData(None, meta, 'default')

	def test_initialize_with_metric_with_feature_default(self):
		"""Initialize ExperimentData with metrics=DF, features='default'"""
		# metrics,meta = generate_random_data()
		D = ExperimentData(self.metrics, self.metadata, 'default')

	def test_initialize_with_metric_with_feature_list(self):
		"""Initialize ExperimentData with metrics=DF, features=list"""
		metrics, meta = generate_random_data()
		D = ExperimentData(metrics, meta, [])
		D = ExperimentData(metrics, meta, [4, 6])

	def test_initialize_with_metric_with_feature_df(self):
		"""Initialize ExperimentData with metrics=DF, features=DF"""
		metrics, meta = generate_random_data()
		features = deepcopy(metrics)
		D = ExperimentData(metrics, meta, features)

	def test_init_with_aggregated_kpi(self):
		"""Initialize ExperimentData with aggregated KPI data"""
		D = ExperimentData(self.metrics, self.metadata, [4])
		self.assertIsNone(D.kpis_time)

	def test_init_with_time_resolved_kpi(self):
		"""Initialize ExperimentData with time-resolved KPI data"""
		n = 5  # replicate raw data and create synthetic time domain
		metrics_time = self.metrics.loc[np.repeat(self.metrics.index.values, n)]
		metrics_time.loc[:, 'time_since_treatment'] = np.tile(range(5), self.metrics.shape[0])

		D = ExperimentData(metrics_time, self.metadata, [4])
		self.assertIsNotNone(D.kpis_time)
		self.assertEquals(D.kpis.shape[0] * n, D.kpis_time.shape[0])


if __name__ == '__main__':
	unittest.main()



#
__version__ = "0.3.3"


def version_numbers():
	return [int(x) for x in __version__.split('.')]


def git_commit_count():
	"""
  Returns the output of `git rev-list --count HEAD` as an int

  http://programmers.stackexchange.com/a/151558
  """
	print
	'dd'
	commit_count = None

	import subprocess
	output = subprocess.check_output(
		["git", "rev-list", "--count", "HEAD"])
	commit_count = int(output)

	return commit_count


def git_latest_commit():
	""""
  Returns output of `git rev-parse HEAD`

  http://programmers.stackexchange.com/a/151558
  """
	print
	'ee'
	latest_commit = None
	import subprocess
	output = subprocess.check_output(
		["git", "rev-parse", "HEAD"]).strip()

	return output


def version(format_str='{short}'):
	major, minor, patch = version_numbers()

	format_dict = {
		'major': major,
		'minor': minor,
		'patch': patch,
		'short': 'v{:d}.{:d}.{:d}'.format(major, minor, patch),
	}

	format_str = format_str.replace(
		'{commits', '{commit_count').replace(
		'{num_commits', '{commit_count').replace(
		'{last_commit', '{latest_commit').replace(
		'{commit', '{latest_commit').replace(
		'{hash', '{latest_commit').replace(
		'{HEAD', '{latest_commit').replace(
		'{long}', 'v{major:d}.{minor:d}.{patch:d}.{commit_count:d}-r{latest_commit}')

	if ('{commit_count' in format_str) or ('{latest_commit' in format_str):
		format_dict['commit_count'] = git_commit_count()
		format_dict['latest_commit'] = git_latest_commit()

	return format_str.format(**format_dict)


if __name__ == '__main__':
	print
	version('Short Version String: "{short}"')
	print
	version('Full Version String: "{long}"')



__all__ = ["binning", "debugging", "experiment", "experimentdata", "results", "statistics", "util", "version"]

from version import __version__, version

print 'ExpAn core init: {}'.format(version())



"""
ExperimentData module allows us to abstract the data format from the library
such that we can have multiple fetcher modules that only need import this
module, and none of our analysis modules need know anything about the fetchers.

In other words, this is the interface between data fetching and data analysis.

"""

import copy

import numpy as np
import pandas as pd

class ExperimentData(object):
	# TODO: allow definition of the name of 'entity': would be nicer if the index
	# itself maintained the name 'chash' or 'order_number' etc.
	# TODO: explain the mandatory_metadata in the exception raised
	mandatory_metadata = {'experiment', 'source'}
	primary_indices = ['entity', 'variant']
	optional_kpi_indices = ['time_since_treatment']
	known_feature_metrics = {
		'age',
		'zalando_age',
		'customer_age',
		'customer_zalando_age',
		'gender', 'sex',
		'feature',
		'treatment_start_time',
		'orders_existing',
		'orders_prev_year',
		'start_segment',
		# 'business_customer', #these SHOULD be features, but because we have no historical data, their value will depend on when we retrieve the data - i.e. they can change after the experiment.
		# 'corporate_customer',
		# 'special_customer',
		'existing_customer',
		'exposed_customer',
		'clv',
		# I know this is somewhat controversial, but I want to insist on using PCII as KPI and CLV as a feature always meaning 'sum of PCII over lifetime up to treatment start'
	}

	def __init__(self, metrics=None, metadata={}, features='default',
				 deepcopy=False):
		"""
    Want to be able to create results from just a single dataframe.

    Args:
        metrics: data frame that contains either KPI or feature
        metadata: the metadata dict
        features: either 'default', which searches the metrics data frame for predefined feature names
                  or list, which subsets the metrics data frame with the given column indices
                  or data frame, which is feature data frame itself and metrics is either KPI or None
                  or None
        deepcopy: the internal data frames are, by default, shallow copies of the input
                            dataframes: this means the actual data arrays underlying the frames are
                            references to the input. In most use-cases, this is desired (reindexing
                            will not reindex the original etc.) but it may have some edge-case issues.
    """

		self.metadata = metadata or {}

		feature_indices = copy.deepcopy(self.primary_indices)
		kpi_indices = copy.deepcopy(self.primary_indices)
		if metrics is not None:
			kpi_indices += [i for i in self.optional_kpi_indices if i in metrics.columns]

		if metrics is None:
			if not isinstance(features, pd.DataFrame):
				raise ValueError('No metrics provided!')
			else:
				self.kpis = pd.DataFrame(columns=self.primary_indices)
				self.features = features.copy(deep=deepcopy)
				self.variant_names = set(np.unique(self.features.variant))
		else:
			if isinstance(features, pd.DataFrame):
				self.kpis = metrics.copy(deep=deepcopy)
				self.features = features.copy(deep=deepcopy)
				self.variant_names = set(np.unique(self.features.variant))
			elif isinstance(features, list):
				if len(features) == 0:
					self.kpis = metrics.copy(deep=deepcopy)
					self.features = pd.DataFrame(columns=self.primary_indices)
					self.variant_names = set(np.unique(self.kpis.variant))
				else:
					self.kpis = metrics.drop(metrics.columns[features], axis=1)
					primary_idx = [metrics.columns.get_loc(x) for x in self.primary_indices]
					feature_idx = primary_idx + features
					self.features = metrics.iloc[:, feature_idx]
					self.variant_names = set(np.unique(self.features.variant))
			elif features == 'default':
				# TODO: use the detect_features function
				features_present = {m for m in metrics if
									m.lower() in self.known_feature_metrics | set(feature_indices)}
				kpis_present = {m for m in metrics if m.lower() not in self.known_feature_metrics}
				self.features = metrics.loc[:, features_present]
				self.kpis = metrics.loc[:, kpis_present]
				self.variant_names = set(np.unique(self.features.variant))
			elif features is None:
				self.kpis = metrics.copy(deep=deepcopy)
				self.features = pd.DataFrame(columns=self.primary_indices)
				self.variant_names = set(np.unique(self.kpis.variant))
			else:
				raise ValueError('Unknown features argument!')

		# validations
		missing_metadata = set(self.mandatory_metadata) - set(self.metadata)
		if len(missing_metadata) > 0:
			raise KeyError('ExperimentData requires metadata: [' + ','.join(missing_metadata) + ']')

		if 'primary_KPI' in self.metadata and (self.metadata['primary_KPI'] not in self.kpis):
			raise KeyError('ExperimentData requires the primary_KPI (\'{}\') to be present in the KPIs.'.format(
				self.metadata['primary_KPI']))

		if len(self.variant_names) < 2:
			raise KeyError('Less than 2 variants found!')

		self.features.set_index(list(feature_indices), inplace=True)
		self.kpis.set_index(list(kpi_indices), inplace=True)

		# check if time domain exists
		if 'time_since_treatment' in self.kpis.index.names:
			self.kpis_time = self.kpis
			# discard NaNs in the time dimension
			self.kpis_time = self.kpis_time[(~self.kpis_time.reset_index('time_since_treatment').time_since_treatment.isnull()).tolist()]
			# NOTE: for most current implemented KPIs, the sum aggregation is
			# appropriate
			self.kpis = self.kpis_time.groupby(level=['entity', 'variant']).sum()
		else:
			self.kpis_time = None

	@property
	def feature_names(self):
		"""List of features"""
		return set([] if self.features is None else self.features.columns)

	@property
	def kpi_names(self):
		"""List of KPIs"""
		return set([] if self.kpis is None else self.kpis.columns)

	# @property
	# def variant_names(self):
	# 	"""List of Variants"""
	# 	return set([] if self.features is None else set(self.features.index.get_level_values('variant')))

	@property
	def metric_names(self):
		"""List of metrics (KPIs and features)"""
		return self.feature_names.union(self.kpi_names)

	def __str__(self):
		return '{} \'{}\' with {:d} features and {:d} KPIs (primary: \'{}\'), {:d} entities'.format(
			self.__class__.__name__,
			self.metadata['experiment'],
			len(self.feature_names),
			len(self.kpi_names),
			self.metadata.get('primary_KPI', 'undefined'),
			self.features.index.nunique()  # TODO: should we explicitly count unique entities?
		)

	def __repr__(self):
		# TODO: improve this
		return 'ExperimentData(\nkpis={}\nfeatures={}\nmetadata={}'.format(
			repr(self.kpis),
			repr(self.features),
			repr(self.metadata),
		)

	@property
	def metrics(self):
		"""
    Simply joins the KPIs and the features.

    TODO: it may well be worth investigating caching this result because the
    features and kpis will rarely change, and accessing them in this way is likely to be common.
    """
		if 'time_since_treatment' in self.kpis.index.names:
			return self.kpis.reset_index('time_since_treatment').join(self.features).set_index('time_since_treatment',
																							   append=True)
		else:
			return self.kpis.join(self.features)

	def __getitem__(self, key):
		"""
    Allows indexing the ExperimentData directly as though it were a DataFrame
    composed of KPIs and Features.
    """
		return self.metrics.__getitem__(key)

	# Q: is it possible to pass a whole lot of functions to
	# a member variable without specifying each?
	def feature_boxplot(self, feature, kpi, **kwargs):
		self.metrics.set_index(feature, append=True).unstack(level=['variant', feature])[kpi].boxplot(**kwargs)


def detect_features(metrics):
	"""
  Automatically detect which of the metrics are features.
  """
	from warnings import warn

	if 'time_since_treatment' in metrics:
		# TODO: test this!
		# assuming that time is the only optional extra index for kpis...
		nunique = metrics.groupby(ExperimentData.primary_indices).agg(
			lambda x: len(x.unique())).max()
		features_present = (nunique == 1)
	# TODO: drop time dimension from features (i.e. take first value)
	else:
		features_present = {m for m in metrics if m.lower() in
							ExperimentData.known_feature_metrics}

	warn('not tested')

	return features_present


if __name__ == '__main__':
	from tests.tests_core.test_data import generate_random_data

	np.random.seed(0)
	metrics, meta = generate_random_data()
	D = ExperimentData(metrics, meta, 'default')



#!/usr/bin/env python

import re
from pip.req import parse_requirements
from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext as _build_ext

try:
	install_reqs = parse_requirements('requirements.txt', session=False)
	requirements = [str(ir.req) for ir in install_reqs]
except:
	requirements = []

with open('README.rst') as readme_file:
	readme = readme_file.read()

with open('HISTORY.rst') as history_file:
	history = history_file.read()

with open('expan/core/version.py', 'r') as fd:
	version = re.search(
		r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
		fd.read(),
		re.MULTILINE).group(1)

if not version:
	raise RuntimeError('Cannot find version information')

test_requirements = [
	'pytest'
]


class build_ext(_build_ext):
	def finalize_options(self):
		_build_ext.finalize_options(self)
		# Prevent numpy from thinking it is still in its setup process:
		__builtins__.__NUMPY_SETUP__ = False
		import numpy
		self.include_dirs.append(numpy.get_include())


setup(
	name='expan',
	version=version,
	description="Experiment Analysis Library",
	long_description=readme + '\n\n' + history,
	author="Zalando SE",
	author_email='octopus@zalando.de',
	url='https://github.com/zalando/expan',
	packages=find_packages(),
	package_dir={'expan': 'expan'},
	include_package_data=True,
	install_requires=requirements,
	cmdclass={'build_ext': build_ext},
	setup_requires=['numpy'],
	license="MIT",
	zip_safe=False,
	keywords='expan',
	entry_points={
        'console_scripts': [
            'expan = expan.cli.cli:main'
        ]
    },
	classifiers=[
		'Development Status :: 2 - Pre-Alpha',
		'Intended Audience :: Developers',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Natural Language :: English',
		# "Programming Language :: Python :: 2",
		# 'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7',
		# 'Programming Language :: Python :: 3',
		# 'Programming Language :: Python :: 3.3',
		# 'Programming Language :: Python :: 3.4',
		# 'Programming Language :: Python :: 3.5',
	],
	test_suite='tests',
	tests_require=test_requirements
)



import numpy as np
import pandas as pd


def scale_range(x, new_min=0.0, new_max=1.0, old_min=None, old_max=None, squash_outside_range=True, squash_inf=False, ):
	"""
  Scales a sequence to fit within a new range.

  If squash_inf is set, then infinite values will take on the
  extremes of the new range (as opposed to staying infinite).

  NB: Infinity in the input is disregarded in the construction of
  the scale of the mapping.

  >>> scale_range([1,3,5])
  array([ 0. ,  0.5,  1. ])

  >>> scale_range([1,2,3,4,5])
  array([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ])

  >>> scale_range([1,3,5, np.inf])
  array([ 0. ,  0.5,  1. ,  inf])

  >>> scale_range([1,3,5, -np.inf])
  array([ 0. ,  0.5,  1. , -inf])

  >>> scale_range([1,3,5, -np.inf], squash_inf=True)
  array([ 0. ,  0.5,  1. ,  0. ])

  >>> scale_range([1,3,5, np.inf], squash_inf=True)
  array([ 0. ,  0.5,  1. ,  1. ])

  >>> scale_range([1,3,5], new_min=0.5)
  array([ 0.5 ,  0.75,  1.  ])

  >>> scale_range([1,3,5], old_min=1, old_max=4)
  array([ 0.        ,  0.66666667,  1.        ])

  >>> scale_range([5], old_max=4)
  array([ 1.])

  """
	if squash_inf and not squash_outside_range:
		# TODO: "warn" is an unresolved reference
		warn(ValueError(
			'Makes no sense to squash infinity but not other numbers outside the source range. Will squash all outside range.'))
		squash_outside_range = True

	if isinstance(x, list):
		x = np.array(x)

	if old_min is None:
		old_min = np.min(x[~np.isinf(x)])
	if old_max is None:
		old_max = np.max(x[~np.isinf(x)])
	old_range = old_max - old_min
	new_max = float(new_max)
	new_min = float(new_min)
	new_range = new_max - new_min

	retval = (new_range * (x - old_min) / old_range) + new_min
	if squash_inf:
		retval[np.isinf(x) & (x > 0)] = new_max
		retval[np.isinf(x) & (x < 0)] = new_min

	if squash_outside_range:
		retval[~np.isinf(x) & (x > old_max)] = new_max
		retval[~np.isinf(x) & (x < old_min)] = new_min

	return retval


def reindex(df, axis=0):
	"""
  partial fulfilment of:
  https://github.com/pydata/pandas/issues/2770

  TODO: test.
  TODO: incorporate in pandas in drop() call and issue pull request
  """

	if axis not in [0, 1, 'index', 'columns']:
		raise NotImplementedError('only index and columns can be selected for axis')

	axis = 0 if axis in [0, 'index'] else 1
	axis_index = df.index if axis == 0 else df.columns  # TODO: fix for panels?

	new_index = pd.MultiIndex.from_tuples(axis_index.values)

	if axis == 0:
		df.index = new_index
	elif axis == 1:
		df.columns = new_index

	return df


if __name__ == '__main__':
	import doctest

	doctest.testmod()



# fileencoding: utf8
import os
import unittest
import warnings

import numpy as np
import pandas as pd

import expan.core.results as r
from expan.core.experiment import Experiment
from tests.tests_core.test_data import generate_random_data
from expan.core.results import prob_uplift_over_zero_single_metric

reload(r)

data_dir = os.getcwd() + '/tests/tests_core/'  # TODO: adjust depending on where we're called from?


def generate_random_results():
	# TODO
	pass


def load_example_results():
	"""
	This just loads example data so that we need always generate random stuff
	in order to test.

	Also demonstrates loading of HDF5 into expan

	Returns Results object.
	"""

	example_fname = 'example_results.h5'
	example_fpath = os.path.join(data_dir, example_fname)

	return r.from_hdf(example_fpath)

class ResultsTestCase(unittest.TestCase):
	"""
	Defines the setUp() and tearDown() functions for the results test cases.
	"""

	def setUp(self):
		"""
	    Load the needed datasets for all TestCases and set the random
	    seed so that randomized algorithms show deterministic behaviour.
	    """
		np.random.seed(0)
		self.data = Experiment('B', *generate_random_data())
		# Create time column. TODO: Do this nicer
		self.data.kpis['time_since_treatment'] = \
			self.data.features['treatment_start_time']
		# Make time part of index
		self.data.kpis.set_index('time_since_treatment', append=True, inplace=True)

	def tearDown(self):
		"""
	    Clean up after the test
	    """
		# TODO: find out if we have to remove data manually
		pass


class ResultsClassTestCase(ResultsTestCase):
	def testExampleResults(self):
		h5py_available = False
		import imp
		try:
			imp.find_module('h5py')
			imp.find_module('tables')
			h5py_available = True
		except Exception:
			warnings.warn(
				"""Could not import h5py or tables module. Skipping
		        testExampleResults(). Please make sure that you have the h5py
		        and tables packages installed."""
			)

		if h5py_available:
			#aa = load_example_results()
			warnings.warn("No data for h5 loading available... skipping tests of example h5 data")

	def test_relative_uplift_delta(self):
		"""Check if the calculation of relative uplift for delta results is
	    correct.
	    """
		res = self.data.delta()
		df = res.relative_uplift('delta', 'normal_same')
		np.testing.assert_almost_equal(df, np.array([[-4.219601, 0]]), decimal=5)

	def test_prob_uplift_over_zero_single_metric(self):
		"""Check if the conversion from confidence intervals to probability is correct for one metric."""
		res = self.data.delta(kpi_subset=['normal_same'])
		#df = prob_uplift_over_zero_single_metric(res.df, self.data.baseline_variant)
		np.testing.assert_almost_equal(res.df.loc[pd.IndexSlice[:,:,:,'prob_uplift_over_0'], 'value'],
									   np.array([[0.946519, np.nan]]), decimal=5)

	def test_prob_uplift_over_zero_multiple_metric(self):
		"""Check if the conversion from confidence intervals to probability is correct for multiple metrics."""
		res = self.data.delta(kpi_subset=['normal_same','normal_shifted'])
		#res.calculate_prob_uplift_over_zero()
		np.testing.assert_almost_equal(res.df.loc[pd.IndexSlice[:,:,:,'prob_uplift_over_0'], 'value'],
									   np.array([[0.946519,np.nan],[0,np.nan]]), decimal=5)

if __name__ == '__main__':
	#unittest.main()
	np.random.seed(0)
	exp = Experiment('B', *generate_random_data())
	res = exp.delta(['normal_shifted'])



import os
import unittest
from os.path import dirname, join, realpath

import simplejson as json

import expan.data.csv_fetcher as csv_fetcher
import tests.tests_core.test_data as td

__location__ = realpath(join(os.getcwd(), dirname(__file__)))

TEST_FOLDER = __location__ + 'test_folder'


class CsvFetcherTestCase(unittest.TestCase):
	def setUp(self):

		# create test folder
		if not os.path.exists(TEST_FOLDER):
			os.makedirs(TEST_FOLDER)

		# generate metrics and metadata
		(metrics, metadata) = td.generate_random_data()

		# save metrics to .csv.gz file in test folder
		metrics.to_csv(path_or_buf=TEST_FOLDER + '/metrics.csv.gz', compression='gzip')

		# save metadata to .json file in test folder
		with open(TEST_FOLDER + '/metadata.json', 'w') as f:
			json.dump(metadata, f)

	def tearDown(self):

		# remove all test files and test folder
		for root, dirs, files in os.walk(TEST_FOLDER, topdown=False):
			for name in files:
				os.remove(os.path.join(root, name))
			for name in dirs:
				os.rmdir(os.path.join(root, name))
		os.rmdir(TEST_FOLDER)

	def test_csv_fetcher(self):
		# should work:
		csv_fetcher.get_data(TEST_FOLDER)

		# should not work:
		with self.assertRaises(AssertionError):
			csv_fetcher.get_data(__location__ + '/../')



"""
CSV fetcher
Expects as input a folder containing the following files:
 - one .csv or .csv.gz with 'metrics' in the filename
 - one .txt containing 'metadata' in the filename

Opens the files and uses them to create an ExperimentData object which it then returns.
"""

from os import listdir
from os.path import isfile, join

import simplejson as json

from expan.core.experimentdata import *


def get_data(folder_path):
	files = [f for f in listdir(folder_path) if isfile(join(folder_path, f))]

	try:
		assert ('metrics' in '-'.join(files))
		assert ('metadata' in '-'.join(files))

		metrics = metadata = None

		for f in files:

			if 'metrics' in f:
				try:
					metrics = pd.read_csv(folder_path + '/' + f)
				except Exception as e:
					print
					e

			elif 'metadata' in f:
				try:
					with open(folder_path + '/' + f, 'r') as input_json:
						metadata = json.load(input_json)
				except ValueError as e:
					print
					e
					raise

		return ExperimentData(metrics=metrics, metadata=metadata)

	except AssertionError as e:
		print
		e
		raise
